"""Compute functions for profiles.

Notes
-----
Some quantities require additional work to compute at the magnetic axis.
A Python lambda function is used to lazily compute the magnetic axis limits
of these quantities. These lambda functions are evaluated only when the
computational grid has a node on the magnetic axis to avoid potentially
expensive computations.
"""

from scipy.constants import elementary_charge, mu_0

from desc.backend import cond, jnp

from .data_index import register_compute_fun
from .utils import cumtrapz, dot, surface_averages, surface_integrals


@register_compute_fun(
    name="Psi",
    label="\\Psi",
    units="Wb",
    units_long="Webers",
    description="Toroidal flux",
    dim=1,
    params=["Psi"],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["rho"],
)
def _Psi(params, transforms, profiles, data, **kwargs):
    data["Psi"] = params["Psi"] * data["rho"] ** 2
    return data


@register_compute_fun(
    name="psi",
    label="\\psi = \\Psi / (2 \\pi)",
    units="Wb",
    units_long="Webers",
    description="Toroidal flux (normalized by 2pi)",
    dim=1,
    params=["Psi"],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["rho"],
)
def _psi(params, transforms, profiles, data, **kwargs):
    data["psi"] = params["Psi"] * data["rho"] ** 2 / (2 * jnp.pi)
    return data


@register_compute_fun(
    name="psi_r",
    label="\\partial_{\\rho} \\psi = \\partial_{\\rho} \\Psi / (2 \\pi)",
    units="Wb",
    units_long="Webers",
    description="Toroidal flux (normalized by 2pi), first radial derivative",
    dim=1,
    params=["Psi"],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["rho"],
)
def _psi_r(params, transforms, profiles, data, **kwargs):
    data["psi_r"] = params["Psi"] * data["rho"] / jnp.pi
    return data


@register_compute_fun(
    name="psi_rr",
    label="\\partial_{\\rho\\rho} \\psi = \\partial_{\\rho\\rho} \\Psi / (2 \\pi)",
    units="Wb",
    units_long="Webers",
    description="Toroidal flux (normalized by 2pi), second radial derivative",
    dim=1,
    params=["Psi"],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["rho"],
)
def _psi_rr(params, transforms, profiles, data, **kwargs):
    data["psi_rr"] = params["Psi"] * jnp.ones_like(data["rho"]) / jnp.pi
    return data


@register_compute_fun(
    name="psi_rrr",
    label="\\partial_{\\rho\\rho\\rho} \\psi = \\partial_{\\rho\\rho\\rho} \\Psi / "
    + "(2 \\pi)",
    units="Wb",
    units_long="Webers",
    description="Toroidal flux (normalized by 2pi), third radial derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["0"],
)
def _psi_rrr(params, transforms, profiles, data, **kwargs):
    data["psi_rrr"] = data["0"]
    return data


@register_compute_fun(
    name="chi_r",
    label="\\partial_{\\rho} \\chi",
    units="Wb",
    units_long="Webers",
    description="Poloidal flux (normalized by 2pi), first radial derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["psi_r", "iota"],
)
def _chi_r(params, transforms, profiles, data, **kwargs):
    data["chi_r"] = data["psi_r"] * data["iota"]
    return data


@register_compute_fun(
    name="chi",
    label="\\chi",
    units="Wb",
    units_long="Webers",
    description="Poloidal flux (normalized by 2pi)",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["chi_r", "rho"],
)
def _chi(params, transforms, profiles, data, **kwargs):
    chi_r = transforms["grid"].compress(data["chi_r"])
    chi = cumtrapz(chi_r, transforms["grid"].compress(data["rho"]), initial=0)
    data["chi"] = transforms["grid"].expand(chi)
    return data


@register_compute_fun(
    name="Te",
    label="T_e",
    units="eV",
    units_long="electron-Volts",
    description="Electron temperature",
    dim=1,
    params=["Te_l"],
    transforms={"grid": []},
    profiles=["electron_temperature"],
    coordinates="r",
    data=["0"],
)
def _Te(params, transforms, profiles, data, **kwargs):
    if profiles["electron_temperature"] is not None:
        data["Te"] = profiles["electron_temperature"].compute(
            transforms["grid"], params["Te_l"], dr=0
        )
    else:
        data["Te"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="Te_r",
    label="\\partial_{\\rho} T_e",
    units="eV",
    units_long="electron-Volts",
    description="Electron temperature, first radial derivative",
    dim=1,
    params=["Te_l"],
    transforms={"grid": []},
    profiles=["electron_temperature"],
    coordinates="r",
    data=["0"],
)
def _Te_r(params, transforms, profiles, data, **kwargs):
    if profiles["electron_temperature"] is not None:
        data["Te_r"] = profiles["electron_temperature"].compute(
            transforms["grid"], params["Te_l"], dr=1
        )
    else:
        data["Te_r"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="Te_rr",
    label="\\partial_{\\rho \\rho} T_e",
    units="eV",
    units_long="electron-Volts",
    description="Electron temperature, second radial derivative",
    dim=1,
    params=["Te_l"],
    transforms={"grid": []},
    profiles=["electron_temperature"],
    coordinates="r",
    data=["0"],
)
def _Te_rr(params, transforms, profiles, data, **kwargs):
    if profiles["electron_temperature"] is not None:
        data["Te_rr"] = profiles["electron_temperature"].compute(
            transforms["grid"], params["Te_l"], dr=2
        )
    else:
        data["Te_rr"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="ne",
    label="n_e",
    units="m^{-3}",
    units_long="1 / cubic meters",
    description="Electron density",
    dim=1,
    params=["ne_l"],
    transforms={"grid": []},
    profiles=["electron_density"],
    coordinates="r",
    data=["0"],
)
def _ne(params, transforms, profiles, data, **kwargs):
    if profiles["electron_density"] is not None:
        data["ne"] = profiles["electron_density"].compute(
            transforms["grid"], params["ne_l"], dr=0
        )
    else:
        data["ne"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="ne_r",
    label="\\partial_{\\rho} n_e",
    units="m^{-3}",
    units_long="1 / cubic meters",
    description="Electron density, first radial derivative",
    dim=1,
    params=["ne_l"],
    transforms={"grid": []},
    profiles=["electron_density"],
    coordinates="r",
    data=["0"],
)
def _ne_r(params, transforms, profiles, data, **kwargs):
    if profiles["electron_density"] is not None:
        data["ne_r"] = profiles["electron_density"].compute(
            transforms["grid"], params["ne_l"], dr=1
        )
    else:
        data["ne_r"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="ne_rr",
    label="\\partial_{\\rho \\rho} n_e",
    units="m^{-3}",
    units_long="1 / cubic meters",
    description="Electron density, second radial derivative",
    dim=1,
    params=["ne_l"],
    transforms={"grid": []},
    profiles=["electron_density"],
    coordinates="r",
    data=["0"],
)
def _ne_rr(params, transforms, profiles, data, **kwargs):
    if profiles["electron_density"] is not None:
        data["ne_rr"] = profiles["electron_density"].compute(
            transforms["grid"], params["ne_l"], dr=2
        )
    else:
        data["ne_rr"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="Ti",
    label="T_i",
    units="eV",
    units_long="electron-Volts",
    description="Ion temperature",
    dim=1,
    params=["Ti_l"],
    transforms={"grid": []},
    profiles=["ion_temperature"],
    coordinates="r",
    data=["0"],
)
def _Ti(params, transforms, profiles, data, **kwargs):
    if profiles["ion_temperature"] is not None:
        data["Ti"] = profiles["ion_temperature"].compute(
            transforms["grid"], params["Ti_l"], dr=0
        )
    else:
        data["Ti"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="Ti_r",
    label="\\partial_{\\rho} T_i",
    units="eV",
    units_long="electron-Volts",
    description="Ion temperature, first radial derivative",
    dim=1,
    params=["Ti_l"],
    transforms={"grid": []},
    profiles=["ion_temperature"],
    coordinates="r",
    data=["0"],
)
def _Ti_r(params, transforms, profiles, data, **kwargs):
    if profiles["ion_temperature"] is not None:
        data["Ti_r"] = profiles["ion_temperature"].compute(
            transforms["grid"], params["Ti_l"], dr=1
        )
    else:
        data["Ti_r"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="Ti_rr",
    label="\\partial_{\\rho \\rho} T_i",
    units="eV",
    units_long="electron-Volts",
    description="Ion temperature, second radial derivative",
    dim=1,
    params=["Ti_l"],
    transforms={"grid": []},
    profiles=["ion_temperature"],
    coordinates="r",
    data=["0"],
)
def _Ti_rr(params, transforms, profiles, data, **kwargs):
    if profiles["ion_temperature"] is not None:
        data["Ti_rr"] = profiles["ion_temperature"].compute(
            transforms["grid"], params["Ti_l"], dr=2
        )
    else:
        data["Ti_rr"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="Zeff",
    label="Z_{eff}",
    units="~",
    units_long="None",
    description="Effective atomic number",
    dim=1,
    params=["Zeff_l"],
    transforms={"grid": []},
    profiles=["atomic_number"],
    coordinates="r",
    data=["0"],
)
def _Zeff(params, transforms, profiles, data, **kwargs):
    if profiles["atomic_number"] is not None:
        data["Zeff"] = profiles["atomic_number"].compute(
            transforms["grid"], params["Zeff_l"], dr=0
        )
    else:
        data["Zeff"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="Zeff_r",
    label="\\partial_{\\rho} Z_{eff}",
    units="~",
    units_long="None",
    description="Effective atomic number, first radial derivative",
    dim=1,
    params=["Zeff_l"],
    transforms={"grid": []},
    profiles=["atomic_number"],
    coordinates="r",
    data=["0"],
)
def _Zeff_r(params, transforms, profiles, data, **kwargs):
    if profiles["atomic_number"] is not None:
        data["Zeff_r"] = profiles["atomic_number"].compute(
            transforms["grid"], params["Zeff_l"], dr=1
        )
    else:
        data["Zeff_r"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="p",
    label="p",
    units="Pa",
    units_long="Pascals",
    description="Pressure",
    dim=1,
    params=["p_l"],
    transforms={"grid": []},
    profiles=["pressure"],
    coordinates="r",
    data=["Te", "ne", "Ti", "Zeff"],
)
def _p(params, transforms, profiles, data, **kwargs):
    if profiles["pressure"] is not None:
        data["p"] = profiles["pressure"].compute(
            transforms["grid"], params["p_l"], dr=0
        )
    else:
        data["p"] = elementary_charge * (
            data["ne"] * data["Te"] + data["Ti"] * data["ne"] / data["Zeff"]
        )
    return data


@register_compute_fun(
    name="p_r",
    label="\\partial_{\\rho} p",
    units="Pa",
    units_long="Pascals",
    description="Pressure, first radial derivative",
    dim=1,
    params=["p_l"],
    transforms={"grid": []},
    profiles=["pressure"],
    coordinates="r",
    data=["Te", "Te_r", "ne", "ne_r", "Ti", "Ti_r", "Zeff", "Zeff_r"],
)
def _p_r(params, transforms, profiles, data, **kwargs):
    if profiles["pressure"] is not None:
        data["p_r"] = profiles["pressure"].compute(
            transforms["grid"], params["p_l"], dr=1
        )
    else:
        data["p_r"] = elementary_charge * (
            data["ne_r"] * data["Te"]
            + data["ne"] * data["Te_r"]
            + data["Ti_r"] * data["ne"] / data["Zeff"]
            + data["Ti"] * data["ne_r"] / data["Zeff"]
            - data["Ti"] * data["ne"] * data["Zeff_r"] / data["Zeff"] ** 2
        )
    return data


@register_compute_fun(
    name="p_t",
    label="\\partial_{\\theta} p",
    units="Pa",
    units_long="Pascals",
    description="Pressure, first poloidal derivative",
    dim=1,
    params=["p_l"],
    transforms={"grid": []},
    profiles=["pressure"],
    coordinates="rtz",
    data=[],
)
def _p_t(params, transforms, profiles, data, **kwargs):
    data["p_t"] = profiles["pressure"].compute(transforms["grid"], params["p_l"], dt=1)
    return data


@register_compute_fun(
    name="p_z",
    label="\\partial_{\\zeta} p",
    units="Pa",
    units_long="Pascals",
    description="Pressure, first toroidal derivative",
    dim=1,
    params=["p_l"],
    transforms={"grid": []},
    profiles=["pressure"],
    coordinates="rtz",
    data=[],
)
def _p_z(params, transforms, profiles, data, **kwargs):
    data["p_z"] = profiles["pressure"].compute(transforms["grid"], params["p_l"], dz=1)
    return data


@register_compute_fun(
    name="grad(p)",
    label="\\nabla p",
    units="N \\cdot m^{-3}",
    units_long="Newtons / cubic meter",
    description="Pressure gradient",
    dim=3,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["p_r", "p_t", "p_z", "e^rho", "e^theta", "e^zeta"],
)
def _gradp(params, transforms, profiles, data, **kwargs):
    data["grad(p)"] = (
        data["p_r"] * data["e^rho"].T
        # e^theta is blows up at the axis but the limit should go to zero
        # if pressure is analytic
        + jnp.where(data["p_t"] == 0, 0, data["p_t"] * data["e^theta"].T)
        + data["p_z"] * data["e^zeta"].T
    ).T
    return data


@register_compute_fun(
    name="|grad(p)|^2",
    label="|\\nabla p|^{2}",
    units="N^2 \\cdot m^{-6}",
    units_long="Newtons per cubic meter squared",
    description="Magnitude of pressure gradient squared",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["grad(p)"],
)
def _gradp_mag2(params, transforms, profiles, data, **kwargs):
    data["|grad(p)|^2"] = dot(data["grad(p)"], data["grad(p)"])
    return data


@register_compute_fun(
    name="|grad(p)|",
    label="|\\nabla p|",
    units="N \\cdot m^{-3}",
    units_long="Newtons per cubic meter",
    description="Magnitude of pressure gradient",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="rtz",
    data=["|grad(p)|^2"],
)
def _gradp_mag(params, transforms, profiles, data, **kwargs):
    data["|grad(p)|"] = jnp.sqrt(data["|grad(p)|^2"])
    return data


@register_compute_fun(
    name="<|grad(p)|>_vol",
    label="\\langle |\\nabla p| \\rangle_{vol}",
    units="N \\cdot m^{-3}",
    units_long="Newtons per cubic meter",
    description="Volume average of magnitude of pressure gradient",
    dim=0,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="",
    data=["|grad(p)|", "sqrt(g)", "V"],
)
def _gradp_mag_vol(params, transforms, profiles, data, **kwargs):
    data["<|grad(p)|>_vol"] = (
        jnp.sum(data["|grad(p)|"] * data["sqrt(g)"] * transforms["grid"].weights)
        / data["V"]
    )
    return data


@register_compute_fun(
    name="beta_a",
    label="\\beta_a = \\mu_0 (p_{||} - p_{\\perp})/B^2",
    units="~",
    units_long="None",
    description="Pressure anisotropy",
    dim=1,
    params=["a_lmn"],
    transforms={"grid": []},
    profiles=["anisotropy"],
    coordinates="rtz",
    data=["0"],
)
def _beta_a(params, transforms, profiles, data, **kwargs):
    if profiles["anisotropy"] is not None:
        data["beta_a"] = profiles["anisotropy"].compute(
            transforms["grid"], params["a_lmn"], dr=0
        )
    else:
        data["beta_a"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="beta_a_r",
    label="\\partial_{\\rho} \\beta_a = \\mu_0 (p_{||} - p_{\\perp})/B^2",
    units="~",
    units_long="None",
    description="Pressure anisotropy, first radial derivative",
    dim=1,
    params=["a_lmn"],
    transforms={"grid": []},
    profiles=["anisotropy"],
    coordinates="rtz",
    data=["0"],
)
def _beta_a_r(params, transforms, profiles, data, **kwargs):
    if profiles["anisotropy"] is not None:
        data["beta_a_r"] = profiles["anisotropy"].compute(
            transforms["grid"], params["a_lmn"], dr=1
        )
    else:
        data["beta_a_r"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="beta_a_t",
    label="\\partial_{\\theta} \\beta_a = \\mu_0 (p_{||} - p_{\\perp})/B^2",
    units="~",
    units_long="None",
    description="Pressure anisotropy, first poloidal derivative",
    dim=1,
    params=["a_lmn"],
    transforms={"grid": []},
    profiles=["anisotropy"],
    coordinates="rtz",
    data=["0"],
)
def _beta_a_t(params, transforms, profiles, data, **kwargs):
    if profiles["anisotropy"] is not None:
        data["beta_a_t"] = profiles["anisotropy"].compute(
            transforms["grid"], params["a_lmn"], dt=1
        )
    else:
        data["beta_a_t"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="beta_a_z",
    label="\\partial_{\\zeta} \\beta_a = \\mu_0 (p_{||} - p_{\\perp})/B^2",
    units="~",
    units_long="None",
    description="Pressure anisotropy, first toroidal derivative",
    dim=1,
    params=["a_lmn"],
    transforms={"grid": []},
    profiles=["anisotropy"],
    coordinates="rtz",
    data=["0"],
)
def _beta_a_z(params, transforms, profiles, data, **kwargs):
    if profiles["anisotropy"] is not None:
        data["beta_a_z"] = profiles["anisotropy"].compute(
            transforms["grid"], params["a_lmn"], dz=1
        )
    else:
        data["beta_a_z"] = jnp.nan * data["0"]
    return data


@register_compute_fun(
    name="grad(beta_a)",
    label="\\nabla \\beta_a = \\nabla \\mu_0 (p_{||} - p_{\\perp})/B^2",
    units="m^{-1}",
    units_long="Inverse meters",
    description="Pressure anisotropy gradient",
    dim=3,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="rtz",
    data=["beta_a_r", "beta_a_t", "beta_a_z", "e^rho", "e^theta", "e^zeta"],
)
def _gradbeta_a(params, transforms, profiles, data, **kwargs):
    data["grad(beta_a)"] = (
        data["beta_a_r"] * data["e^rho"].T
        + data["beta_a_t"] * data["e^theta"].T
        + data["beta_a_z"] * data["e^zeta"].T
    ).T
    return data


@register_compute_fun(
    name="iota",
    label="\\iota",
    units="~",
    units_long="None",
    description="Rotational transform (normalized by 2pi)",
    dim=1,
    params=["i_l"],
    transforms={"grid": []},
    profiles=["iota", "current"],
    coordinates="r",
    data=["iota_den", "iota_num"],
    axis_limit_data=["iota_den_r", "iota_num_r"],
)
def _iota(params, transforms, profiles, data, **kwargs):
    if profiles["iota"] is not None:
        data["iota"] = profiles["iota"].compute(transforms["grid"], params["i_l"], dr=0)
    elif profiles["current"] is not None:
        # See the document attached to GitHub pull request #556 for the math.
        data["iota"] = transforms["grid"].replace_at_axis(
            data["iota_num"] / data["iota_den"],
            lambda: data["iota_num_r"] / data["iota_den_r"],
        )
    return data


@register_compute_fun(
    name="iota_r",
    label="\\partial_{\\rho} \\iota",
    units="~",
    units_long="None",
    description="Rotational transform (normalized by 2pi), first radial derivative",
    dim=1,
    params=["i_l"],
    transforms={"grid": []},
    profiles=["iota", "current"],
    coordinates="r",
    data=["iota_den", "iota_den_r", "iota_num", "iota_num_r"],
    axis_limit_data=["iota_den_rr", "iota_num_rr"],
)
def _iota_r(params, transforms, profiles, data, **kwargs):
    if profiles["iota"] is not None:
        data["iota_r"] = profiles["iota"].compute(
            transforms["grid"], params["i_l"], dr=1
        )
    elif profiles["current"] is not None:
        # See the document attached to GitHub pull request #556 for the math.
        data["iota_r"] = transforms["grid"].replace_at_axis(
            (
                data["iota_num_r"] * data["iota_den"]
                - data["iota_num"] * data["iota_den_r"]
            )
            / data["iota_den"] ** 2,
            lambda: (
                data["iota_num_rr"] * data["iota_den_r"]
                - data["iota_num_r"] * data["iota_den_rr"]
            )
            / (2 * data["iota_den_r"] ** 2),
        )
    return data


@register_compute_fun(
    name="iota_rr",
    label="\\partial_{\\rho\\rho} \\iota",
    units="~",
    units_long="None",
    description="Rotational transform (normalized by 2pi), second radial derivative",
    dim=1,
    params=["i_l"],
    transforms={"grid": []},
    profiles=["iota", "current"],
    coordinates="r",
    data=[
        "iota_den",
        "iota_den_r",
        "iota_den_rr",
        "iota_num",
        "iota_num_r",
        "iota_num_rr",
    ],
    axis_limit_data=["iota_den_rrr", "iota_num_rrr"],
)
def _iota_rr(params, transforms, profiles, data, **kwargs):
    if profiles["iota"] is not None:
        data["iota_rr"] = profiles["iota"].compute(
            transforms["grid"], params["i_l"], dr=2
        )
    elif profiles["current"] is not None:
        # See the document attached to GitHub pull request #556 for the math.
        data["iota_rr"] = transforms["grid"].replace_at_axis(
            (
                data["iota_num_rr"] * data["iota_den"] ** 2
                - 2 * data["iota_num_r"] * data["iota_den"] * data["iota_den_r"]
                + 2 * data["iota_num"] * data["iota_den_r"] ** 2
                - data["iota_num"] * data["iota_den"] * data["iota_den_rr"]
            )
            / data["iota_den"] ** 3,
            lambda: (
                2 * data["iota_num_rrr"] * data["iota_den_r"] ** 2
                - 3 * data["iota_num_rr"] * data["iota_den_r"] * data["iota_den_rr"]
                + 3 * data["iota_num_r"] * data["iota_den_rr"] ** 2
                - 2 * data["iota_num_r"] * data["iota_den_r"] * data["iota_den_rrr"]
            )
            / (6 * data["iota_den_r"] ** 3),
        )
    return data


@register_compute_fun(
    name="iota current",
    label="\\iota~\\mathrm{from~current}",
    units="~",
    units_long="None",
    description="Rotational transform (normalized by 2pi), current contribution",
    dim=1,
    params=["i_l"],
    transforms={"grid": []},
    profiles=["iota", "current"],
    coordinates="r",
    data=["iota vacuum", "iota_den", "iota_num current"],
    axis_limit_data=["iota_den_r", "iota_num_r current"],
)
def _iota_current(params, transforms, profiles, data, **kwargs):
    if profiles["iota"] is not None:
        iota = profiles["iota"].compute(transforms["grid"], params["i_l"], dr=0)
        data["iota current"] = iota - data["iota vacuum"]
    elif profiles["current"] is not None:
        data["iota current"] = transforms["grid"].replace_at_axis(
            data["iota_num current"] / data["iota_den"],
            lambda: data["iota_num_r current"] / data["iota_den_r"],
        )
    return data


@register_compute_fun(
    name="iota vacuum",
    label="\\iota~\\mathrm{in~vacuum}",
    units="~",
    units_long="None",
    description="Rotational transform (normalized by 2pi), vacuum contribution",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["iota_den", "iota_num vacuum"],
    axis_limit_data=["iota_den_r", "iota_num_r vacuum"],
)
def _iota_vacuum(params, transforms, profiles, data, **kwargs):
    data["iota vacuum"] = transforms["grid"].replace_at_axis(
        data["iota_num vacuum"] / data["iota_den"],
        lambda: data["iota_num_r vacuum"] / data["iota_den_r"],
    )
    return data


@register_compute_fun(
    name="iota_num current",
    label="\\iota_{\\mathrm{numerator}}~\\mathrm{from~current}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Numerator of rotational transform formula, current contribution",
    dim=1,
    params=["c_l", "i_l"],
    transforms={"grid": []},
    profiles=["current", "iota"],
    coordinates="r",
    data=["psi_r", "iota_den", "iota_num vacuum"],
    axis_limit_data=["psi_rr"],
)
def _iota_num_current(params, transforms, profiles, data, **kwargs):
    """Current contribution to the numerator of rotational transform formula."""
    if profiles["iota"] is not None:
        iota = profiles["iota"].compute(transforms["grid"], params["i_l"], dr=0)
        data["iota_num current"] = iota * data["iota_den"] - data["iota_num vacuum"]
    elif profiles["current"] is not None:
        # 4π^2 I = 4π^2 (mu_0 current / 2π) = 2π mu_0 current
        current = profiles["current"].compute(transforms["grid"], params["c_l"], dr=0)
        current_r = profiles["current"].compute(transforms["grid"], params["c_l"], dr=1)
        data["iota_num current"] = (
            2
            * jnp.pi
            * mu_0
            * transforms["grid"].replace_at_axis(
                current / data["psi_r"],
                lambda: current_r / data["psi_rr"],
            )
        )
    return data


@register_compute_fun(
    name="iota_num vacuum",
    label="\\iota_{\\mathrm{numerator}}~\\mathrm{in~vacuum}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Numerator of rotational transform formula, vacuum contribution",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["lambda_z", "g_tt", "lambda_t", "g_tz", "sqrt(g)"],
    axis_limit_data=["g_tz_r", "sqrt(g)_r"],
)
def _iota_num_vacuum(params, transforms, profiles, data, **kwargs):
    """Vacuum contribution to the numerator of rotational transform formula."""
    iota_num_vacuum = transforms["grid"].replace_at_axis(
        (data["lambda_z"] * data["g_tt"] - (1 + data["lambda_t"]) * data["g_tz"])
        / data["sqrt(g)"],
        lambda: -(1 + data["lambda_t"]) * data["g_tz_r"] / data["sqrt(g)_r"],
    )
    data["iota_num vacuum"] = surface_integrals(transforms["grid"], iota_num_vacuum)
    return data


@register_compute_fun(
    name="iota_num_r current",
    label="\\partial_{\\rho} \\iota_{\\mathrm{numerator}}~\\mathrm{from~current}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Numerator of rotational transform formula, current contribution, "
    + "first radial derivative",
    dim=1,
    params=["c_l", "i_l"],
    transforms={"grid": []},
    profiles=["current", "iota"],
    coordinates="r",
    data=["psi_r", "psi_rr", "iota_den", "iota_den_r", "iota_num", "iota_num_r vacuum"],
)
def _iota_num_r_current(params, transforms, profiles, data, **kwargs):
    if profiles["iota"] is not None:
        iota_r = profiles["iota"].compute(transforms["grid"], params["i_l"], dr=1)
        data["iota_num_r current"] = (
            iota_r * data["iota_den"] ** 2 + data["iota_num"] * data["iota_den_r"]
        ) / data["iota_den"] - data["iota_num_r vacuum"]
    elif profiles["current"] is not None:
        # 4π^2 I = 4π^2 (mu_0 current / 2π) = 2π mu_0 current
        current = profiles["current"].compute(transforms["grid"], params["c_l"], dr=0)
        current_r = profiles["current"].compute(transforms["grid"], params["c_l"], dr=1)
        current_rr = profiles["current"].compute(
            transforms["grid"], params["c_l"], dr=2
        )
        data["iota_num_r current"] = (
            2
            * jnp.pi
            * mu_0
            * transforms["grid"].replace_at_axis(
                (current_r * data["psi_r"] - current * data["psi_rr"])
                / data["psi_r"] ** 2,
                lambda: current_rr / (2 * data["psi_rr"]),
            )
        )
    return data


@register_compute_fun(
    name="iota_num_r vacuum",
    label="\\partial_{\\rho} \\iota_{\\mathrm{numerator}}~\\mathrm{in~vacuum}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Numerator of rotational transform formula, vacuum contribution, "
    + "first radial derivative",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=[
        "lambda_t",
        "lambda_rt",
        "lambda_z",
        "lambda_rz",
        "g_tt",
        "g_tt_r",
        "g_tz",
        "g_tz_r",
        "sqrt(g)",
        "sqrt(g)_r",
    ],
    axis_limit_data=["g_tt_rr", "g_tz_rr", "sqrt(g)_rr"],
)
def _iota_num_r_vacuum(params, transforms, profiles, data, **kwargs):
    iota_num_vacuum = (
        data["lambda_z"] * data["g_tt"] - (1 + data["lambda_t"]) * data["g_tz"]
    ) / data["sqrt(g)"]
    iota_num_r_vacuum = transforms["grid"].replace_at_axis(
        (
            data["lambda_rz"] * data["g_tt"]
            + data["lambda_z"] * data["g_tt_r"]
            - data["lambda_rt"] * data["g_tz"]
            - (1 + data["lambda_t"]) * data["g_tz_r"]
            - iota_num_vacuum * data["sqrt(g)_r"]
        )
        / data["sqrt(g)"],
        lambda: (
            (1 + data["lambda_t"])
            * data["g_tz_r"]
            * data["sqrt(g)_rr"]
            / (2 * data["sqrt(g)_r"] ** 2)
            + (
                data["lambda_z"] * data["g_tt_rr"]
                - 2 * data["lambda_rt"] * data["g_tz_r"]
                - (1 + data["lambda_t"]) * data["g_tz_rr"]
            )
            / (2 * data["sqrt(g)_r"])
        ),
    )
    data["iota_num_r vacuum"] = surface_integrals(transforms["grid"], iota_num_r_vacuum)
    return data


@register_compute_fun(
    name="iota_num",
    label="\\iota_{\\mathrm{numerator}}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Numerator of rotational transform formula",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["iota_num current", "iota_num vacuum"],
)
def _iota_num(params, transforms, profiles, data, **kwargs):
    """Numerator of rotational transform formula.

    Computes 𝛼 + 𝛽 as defined in the document attached to the description
    of GitHub pull request #556. 𝛼 supplements the rotational transform with an
    additional term to account for the enclosed net toroidal current.
    """
    data["iota_num"] = data["iota_num current"] + data["iota_num vacuum"]
    return data


@register_compute_fun(
    name="iota_num_r",
    label="\\partial_{\\rho} \\iota_{\\mathrm{numerator}}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Numerator of rotational transform formula, first radial derivative",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["iota_num_r current", "iota_num_r vacuum"],
)
def _iota_num_r(params, transforms, profiles, data, **kwargs):
    """Numerator of rotational transform formula, first radial derivative.

    Computes d(𝛼+𝛽)/d𝜌 as defined in the document attached to the description
    of GitHub pull request #556. 𝛼 supplements the rotational transform with an
    additional term to account for the enclosed net toroidal current.
    """
    data["iota_num_r"] = data["iota_num_r current"] + data["iota_num_r vacuum"]
    return data


@register_compute_fun(
    name="iota_num_rr",
    label="\\partial_{\\rho\\rho} \\iota_{\\mathrm{numerator}}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Numerator of rotational transform formula, second radial derivative",
    dim=1,
    params=["c_l"],
    transforms={"grid": []},
    profiles=["current", "iota"],
    coordinates="r",
    data=[
        "0",
        "lambda_t",
        "lambda_rt",
        "lambda_rrt",
        "lambda_z",
        "lambda_rz",
        "lambda_rrz",
        "g_tt",
        "g_tt_r",
        "g_tt_rr",
        "g_tz",
        "g_tz_r",
        "g_tz_rr",
        "sqrt(g)",
        "sqrt(g)_r",
        "sqrt(g)_rr",
        "psi_r",
        "psi_rr",
        "psi_rrr",
    ],
    axis_limit_data=["sqrt(g)_rrr", "g_tt_rrr", "g_tz_rrr"],
)
def _iota_num_rr(params, transforms, profiles, data, **kwargs):
    """Numerator of rotational transform formula, second radial derivative.

    Computes d2(𝛼+𝛽)/d𝜌2 as defined in the document attached to the description
    of GitHub pull request #556. 𝛼 supplements the rotational transform with an
    additional term to account for the enclosed net toroidal current.
    """
    if profiles["iota"] is not None:
        data["iota_num_rr"] = jnp.nan * data["0"]
    elif profiles["current"] is not None:
        # 4π^2 I = 4π^2 (mu_0 current / 2π) = 2π mu_0 current
        current = profiles["current"].compute(transforms["grid"], params["c_l"], dr=0)
        current_r = profiles["current"].compute(transforms["grid"], params["c_l"], dr=1)
        current_rr = profiles["current"].compute(
            transforms["grid"], params["c_l"], dr=2
        )
        current_rrr = profiles["current"].compute(
            transforms["grid"], params["c_l"], dr=3
        )
        alpha_rr = (
            jnp.pi
            * mu_0
            * transforms["grid"].replace_at_axis(
                2 * current_rr / data["psi_r"]
                - 4 * current_r * data["psi_rr"] / data["psi_r"] ** 2
                + 2
                * current
                * (2 * data["psi_rr"] ** 2 - data["psi_rrr"] * data["psi_r"])
                / data["psi_r"] ** 3,
                lambda: 2 * current_rrr / (3 * data["psi_rr"])
                - current_rr * data["psi_rrr"] / data["psi_rr"] ** 2
                + current_r * data["psi_rrr"] ** 2 / data["psi_rr"] ** 3,
            )
        )
        beta = (
            data["lambda_z"] * data["g_tt"] - (1 + data["lambda_t"]) * data["g_tz"]
        ) / data["sqrt(g)"]
        beta_r = (
            data["lambda_rz"] * data["g_tt"]
            + data["lambda_z"] * data["g_tt_r"]
            - data["lambda_rt"] * data["g_tz"]
            - (1 + data["lambda_t"]) * data["g_tz_r"]
            - beta * data["sqrt(g)_r"]
        ) / data["sqrt(g)"]
        beta_rr = transforms["grid"].replace_at_axis(
            (
                data["lambda_rrz"] * data["g_tt"]
                + 2 * data["lambda_rz"] * data["g_tt_r"]
                + data["lambda_z"] * data["g_tt_rr"]
                - data["lambda_rrt"] * data["g_tz"]
                - 2 * data["lambda_rt"] * data["g_tz_r"]
                - (1 + data["lambda_t"]) * data["g_tz_rr"]
                - 2 * beta_r * data["sqrt(g)_r"]
                - beta * data["sqrt(g)_rr"]
            )
            / data["sqrt(g)"],
            lambda: (
                2
                * data["sqrt(g)_r"] ** 2
                * (
                    3 * data["g_tt_rr"] * data["lambda_rz"]
                    + data["g_tt_rrr"] * data["lambda_z"]
                    - 3 * data["g_tz_rr"] * data["lambda_rt"]
                    - 3 * data["g_tz_r"] * data["lambda_rrt"]
                    - data["g_tz_rrr"] * (1 + data["lambda_t"])
                )
                + data["sqrt(g)_r"]
                * (
                    3
                    * data["sqrt(g)_rr"]
                    * (
                        2 * data["g_tz_r"] * data["lambda_rt"]
                        - data["g_tt_rr"] * data["lambda_t"]
                        + data["g_tz_rr"] * (1 + data["lambda_t"])
                    )
                    + 2 * data["sqrt(g)_rrr"] * data["g_tz_r"] * (1 + data["lambda_t"])
                )
                - 3 * data["sqrt(g)_rr"] ** 2 * data["g_tz_r"] * (1 + data["lambda_t"])
            )
            / (6 * data["sqrt(g)_r"] ** 3),
        )
        beta_rr = surface_integrals(transforms["grid"], beta_rr)
        data["iota_num_rr"] = alpha_rr + beta_rr
    return data


@register_compute_fun(
    name="iota_num_rrr",
    label="\\partial_{\\rho\\rho\\rho} \\iota_{\\mathrm{numerator}}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Numerator of rotational transform formula, third radial derivative",
    dim=1,
    params=["c_l"],
    transforms={"grid": []},
    profiles=["current", "iota"],
    coordinates="r",
    data=[
        "0",
        "lambda_t",
        "lambda_rt",
        "lambda_rrt",
        "lambda_rrrt",
        "lambda_z",
        "lambda_rz",
        "lambda_rrz",
        "lambda_rrrz",
        "g_tt",
        "g_tt_r",
        "g_tt_rr",
        "g_tt_rrr",
        "g_tz",
        "g_tz_r",
        "g_tz_rr",
        "g_tz_rrr",
        "sqrt(g)",
        "sqrt(g)_r",
        "sqrt(g)_rr",
        "sqrt(g)_rrr",
        "psi_r",
        "psi_rr",
        "psi_rrr",
    ],
)
def _iota_num_rrr(params, transforms, profiles, data, **kwargs):
    """Numerator of rotational transform formula, third radial derivative.

    Computes d3(𝛼+𝛽)/d𝜌3 as defined in the document attached to the description
    of GitHub pull request #556. 𝛼 supplements the rotational transform with an
    additional term to account for the enclosed net toroidal current.
    """
    if profiles["iota"] is not None:
        data["iota_num_rrr"] = jnp.nan * data["0"]
    elif profiles["current"] is not None:
        current = profiles["current"].compute(transforms["grid"], params["c_l"], dr=0)
        current_r = profiles["current"].compute(transforms["grid"], params["c_l"], dr=1)
        current_rr = profiles["current"].compute(
            transforms["grid"], params["c_l"], dr=2
        )
        current_rrr = profiles["current"].compute(
            transforms["grid"], params["c_l"], dr=3
        )
        current_rrrr = profiles["current"].compute(
            transforms["grid"], params["c_l"], dr=4
        )
        # 4π^2 I = 4π^2 (mu_0 current / 2π) = 2π mu_0 current
        alpha_rrr = (
            jnp.pi
            * mu_0
            * transforms["grid"].replace_at_axis(
                2 * current_rrr / data["psi_r"]
                - 6 * current_rr * data["psi_rr"] / data["psi_r"] ** 2
                + 6
                * current_r
                * (
                    2 * data["psi_r"] * data["psi_rr"] ** 2
                    - data["psi_rrr"] * data["psi_r"] ** 2
                )
                / data["psi_r"] ** 4
                + 12
                * current
                * (
                    data["psi_rrr"] * data["psi_rr"] * data["psi_r"]
                    - data["psi_rr"] ** 3
                )
                / data["psi_r"] ** 4,
                lambda: current_rrrr / (2 * data["psi_rr"])
                - current_rrr * data["psi_rrr"] / data["psi_rr"] ** 2
                + 3 * current_rr * data["psi_rrr"] ** 2 / (2 * data["psi_rr"] ** 3)
                - 3 * current_r * data["psi_rrr"] ** 3 / (2 * data["psi_rr"] ** 4),
            )
        )
        beta = (
            data["lambda_z"] * data["g_tt"] - (1 + data["lambda_t"]) * data["g_tz"]
        ) / data["sqrt(g)"]
        beta_r = (
            data["lambda_rz"] * data["g_tt"]
            + data["lambda_z"] * data["g_tt_r"]
            - data["lambda_rt"] * data["g_tz"]
            - (1 + data["lambda_t"]) * data["g_tz_r"]
            - beta * data["sqrt(g)_r"]
        ) / data["sqrt(g)"]
        beta_rr = (
            data["lambda_rrz"] * data["g_tt"]
            + 2 * data["lambda_rz"] * data["g_tt_r"]
            + data["lambda_z"] * data["g_tt_rr"]
            - data["lambda_rrt"] * data["g_tz"]
            - 2 * data["lambda_rt"] * data["g_tz_r"]
            - (1 + data["lambda_t"]) * data["g_tz_rr"]
            - 2 * beta_r * data["sqrt(g)_r"]
            - beta * data["sqrt(g)_rr"]
        ) / data["sqrt(g)"]
        beta_rrr = transforms["grid"].replace_at_axis(
            (
                data["lambda_rrrz"] * data["g_tt"]
                + 3 * data["lambda_rrz"] * data["g_tt_r"]
                + 3 * data["lambda_rz"] * data["g_tt_rr"]
                + data["lambda_z"] * data["g_tt_rrr"]
                - data["lambda_rrrt"] * data["g_tz"]
                - 3 * data["lambda_rrt"] * data["g_tz_r"]
                - 3 * data["lambda_rt"] * data["g_tz_rr"]
                - (1 + data["lambda_t"]) * data["g_tz_rrr"]
                - 3 * beta_rr * data["sqrt(g)_r"]
                - 3 * beta_r * data["sqrt(g)_rr"]
                - beta * data["sqrt(g)_rrr"]
            )
            / data["sqrt(g)"],
            # Todo: axis limit of beta_rrr
            #   Computed with four applications of l’Hôpital’s rule.
            #   Requires sqrt(g)_rrrr and fourth derivatives of basis vectors.
            jnp.nan,
        )
        beta_rrr = surface_integrals(transforms["grid"], beta_rrr)
        # force limit to nan until completed because integration replaces nan with 0
        data["iota_num_rrr"] = alpha_rrr + transforms["grid"].replace_at_axis(
            beta_rrr, jnp.nan
        )
    return data


@register_compute_fun(
    name="iota_den",
    label="\\iota_{\\mathrm{denominator}}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Denominator of rotational transform formula",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["g_tt", "g_tz", "sqrt(g)", "omega_t", "omega_z"],
)
def _iota_den(params, transforms, profiles, data, **kwargs):
    """Denominator of rotational transform formula.

    Computes 𝛾 as defined in the document attached to the description
    of GitHub pull request #556.
    """
    gamma = (
        (1 + data["omega_z"]) * data["g_tt"] - data["omega_t"] * data["g_tz"]
    ) / data["sqrt(g)"]
    # Assumes toroidal stream function behaves such that the magnetic axis limit
    # of gamma is zero (as it would if omega = 0 identically).
    gamma = transforms["grid"].replace_at_axis(
        surface_integrals(transforms["grid"], gamma), 0
    )
    data["iota_den"] = gamma
    return data


@register_compute_fun(
    name="iota_den_r",
    label="\\partial_{\\rho} \\iota_{\\mathrm{denominator}}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Denominator of rotational transform formula, first radial derivative",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=[
        "g_tt",
        "g_tt_r",
        "g_tz",
        "g_tz_r",
        "sqrt(g)",
        "sqrt(g)_r",
        "omega_t",
        "omega_rt",
        "omega_z",
        "omega_rz",
    ],
    axis_limit_data=["sqrt(g)_rr", "g_tt_rr", "g_tz_rr"],
)
def _iota_den_r(params, transforms, profiles, data, **kwargs):
    """Denominator of rotational transform formula, first radial derivative.

    Computes d𝛾/d𝜌 as defined in the document attached to the description
    of GitHub pull request #556.
    """
    gamma = (
        (1 + data["omega_z"]) * data["g_tt"] - data["omega_t"] * data["g_tz"]
    ) / data["sqrt(g)"]
    gamma_r = transforms["grid"].replace_at_axis(
        (
            data["omega_rz"] * data["g_tt"]
            + (1 + data["omega_z"]) * data["g_tt_r"]
            - data["omega_rt"] * data["g_tz"]
            - data["omega_t"] * data["g_tz_r"]
            - gamma * data["sqrt(g)_r"]
        )
        / data["sqrt(g)"],
        lambda: (
            data["omega_t"]
            * data["g_tz_r"]
            * data["sqrt(g)_rr"]
            / (2 * data["sqrt(g)_r"] ** 2)
            + (
                (1 + data["omega_z"]) * data["g_tt_rr"]
                - 2 * data["omega_rt"] * data["g_tz_r"]
                - data["omega_t"] * data["g_tz_rr"]
            )
            / (2 * data["sqrt(g)_r"])
        ),
    )
    gamma_r = surface_integrals(transforms["grid"], gamma_r)
    data["iota_den_r"] = gamma_r
    return data


@register_compute_fun(
    name="iota_den_rr",
    label="\\partial_{\\rho\\rho} \\iota_{\\mathrm{denominator}}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Denominator of rotational transform formula, second radial derivative",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=[
        "g_tt",
        "g_tt_r",
        "g_tt_rr",
        "g_tz",
        "g_tz_r",
        "g_tz_rr",
        "sqrt(g)",
        "sqrt(g)_r",
        "sqrt(g)_rr",
        "omega_t",
        "omega_rt",
        "omega_rrt",
        "omega_z",
        "omega_rz",
        "omega_rrz",
    ],
    axis_limit_data=["sqrt(g)_rrr", "g_tt_rrr", "g_tz_rrr"],
)
def _iota_den_rr(params, transforms, profiles, data, **kwargs):
    """Denominator of rotational transform formula, second radial derivative.

    Computes d2𝛾/d𝜌2 as defined in the document attached to the description
    of GitHub pull request #556.
    """
    gamma = (
        (1 + data["omega_z"]) * data["g_tt"] - data["omega_t"] * data["g_tz"]
    ) / data["sqrt(g)"]
    gamma_r = (
        data["omega_rz"] * data["g_tt"]
        + (1 + data["omega_z"]) * data["g_tt_r"]
        - data["omega_rt"] * data["g_tz"]
        - data["omega_t"] * data["g_tz_r"]
        - gamma * data["sqrt(g)_r"]
    ) / data["sqrt(g)"]
    gamma_rr = transforms["grid"].replace_at_axis(
        (
            data["omega_rrz"] * data["g_tt"]
            + 2 * data["omega_rz"] * data["g_tt_r"]
            + (1 + data["omega_z"]) * data["g_tt_rr"]
            - data["omega_rrt"] * data["g_tz"]
            - 2 * data["omega_rt"] * data["g_tz_r"]
            - data["omega_t"] * data["g_tz_rr"]
            - 2 * gamma_r * data["sqrt(g)_r"]
            - gamma * data["sqrt(g)_rr"]
        )
        / data["sqrt(g)"],
        lambda: (
            2
            * data["sqrt(g)_r"] ** 2
            * (
                3 * data["g_tt_rr"] * data["omega_rz"]
                + data["g_tt_rrr"] * (1 + data["omega_z"])
                - 3 * data["g_tz_rr"] * data["omega_rt"]
                - 3 * data["g_tz_r"] * data["omega_rrt"]
                - data["g_tz_rrr"] * data["omega_t"]
            )
            + data["sqrt(g)_r"]
            * (
                3
                * data["sqrt(g)_rr"]
                * (
                    2 * data["g_tz_r"] * data["omega_rt"]
                    - data["g_tt_rr"] * (1 + data["omega_z"])
                    + data["g_tz_rr"] * data["omega_t"]
                )
                + 2 * data["sqrt(g)_rrr"] * data["g_tz_r"] * data["omega_t"]
            )
            - 3 * data["sqrt(g)_rr"] ** 2 * data["g_tz_r"] * data["omega_t"]
        )
        / (6 * data["sqrt(g)_r"] ** 3),
    )
    gamma_rr = surface_integrals(transforms["grid"], gamma_rr)
    data["iota_den_rr"] = gamma_rr
    return data


@register_compute_fun(
    name="iota_den_rrr",
    label="\\partial_{\\rho\\rho\\rho} \\iota_{\\mathrm{denominator}}",
    units="m^{-1}",
    units_long="inverse meters",
    description="Denominator of rotational transform formula, third radial derivative",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=[
        "g_tt",
        "g_tt_r",
        "g_tt_rr",
        "g_tt_rrr",
        "g_tz",
        "g_tz_r",
        "g_tz_rr",
        "g_tz_rrr",
        "sqrt(g)",
        "sqrt(g)_r",
        "sqrt(g)_rr",
        "sqrt(g)_rrr",
        "omega_t",
        "omega_rt",
        "omega_rrt",
        "omega_rrrt",
        "omega_z",
        "omega_rz",
        "omega_rrz",
        "omega_rrrz",
    ],
)
def _iota_den_rrr(params, transforms, profiles, data, **kwargs):
    """Denominator of rotational transform formula, third radial derivative.

    Computes d3𝛾/d𝜌3 as defined in the document attached to the description
    of GitHub pull request #556.
    """
    gamma = (
        (1 + data["omega_z"]) * data["g_tt"] - data["omega_t"] * data["g_tz"]
    ) / data["sqrt(g)"]
    gamma_r = (
        data["omega_rz"] * data["g_tt"]
        + (1 + data["omega_z"]) * data["g_tt_r"]
        - data["omega_rt"] * data["g_tz"]
        - data["omega_t"] * data["g_tz_r"]
        - gamma * data["sqrt(g)_r"]
    ) / data["sqrt(g)"]
    gamma_rr = (
        data["omega_rrz"] * data["g_tt"]
        + 2 * data["omega_rz"] * data["g_tt_r"]
        + (1 + data["omega_z"]) * data["g_tt_rr"]
        - data["omega_rrt"] * data["g_tz"]
        - 2 * data["omega_rt"] * data["g_tz_r"]
        - data["omega_t"] * data["g_tz_rr"]
        - 2 * gamma_r * data["sqrt(g)_r"]
        - gamma * data["sqrt(g)_rr"]
    ) / data["sqrt(g)"]
    gamma_rrr = transforms["grid"].replace_at_axis(
        (
            data["omega_rrrz"] * data["g_tt"]
            + 3 * data["omega_rrz"] * data["g_tt_r"]
            + 3 * data["omega_rz"] * data["g_tt_rr"]
            + (1 + data["omega_z"]) * data["g_tt_rrr"]
            - data["omega_rrrt"] * data["g_tz"]
            - 3 * data["omega_rrt"] * data["g_tz_r"]
            - 3 * data["omega_rt"] * data["g_tz_rr"]
            - data["omega_t"] * data["g_tz_rrr"]
            - 3 * gamma_rr * data["sqrt(g)_r"]
            - 3 * gamma_r * data["sqrt(g)_rr"]
            - gamma * data["sqrt(g)_rrr"]
        )
        / data["sqrt(g)"],
        # Todo: axis limit
        #   Computed with four applications of l’Hôpital’s rule.
        #   Requires sqrt(g)_rrrr and fourth derivatives of basis vectors.
        jnp.nan,
    )
    gamma_rrr = surface_integrals(transforms["grid"], gamma_rrr)
    # force limit to nan until completed because integration replaces nan with 0
    data["iota_den_rrr"] = transforms["grid"].replace_at_axis(gamma_rrr, jnp.nan)
    return data


@register_compute_fun(
    name="iota_psi",
    label="\\partial_{\\psi} \\iota",
    units="Wb^{-1}",
    units_long="Inverse Webers",
    description="Rotational transform, radial derivative wrt toroidal flux",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["iota_r", "psi_r"],
    axis_limit_data=["iota_rr", "psi_rr"],
)
def _iota_psi(params, transforms, profiles, data, **kwargs):
    # Existence of limit at magnetic axis requires ∂ᵨ iota = 0 at axis.
    # Assume iota may be expanded as an even power series of ρ so that this
    # condition is satisfied.
    data["iota_psi"] = transforms["grid"].replace_at_axis(
        data["iota_r"] / data["psi_r"], lambda: data["iota_rr"] / data["psi_rr"]
    )
    return data


@register_compute_fun(
    name="q",
    label="q = 1/\\iota",
    units="~",
    units_long="None",
    description="Safety factor 'q', inverse of rotational transform.",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["iota"],
)
def _q(params, transforms, profiles, data, **kwargs):
    data["q"] = 1 / data["iota"]
    return data


# TODO: add K(rho,theta,zeta)*grad(rho) term
@register_compute_fun(
    name="I",
    label="I",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Covariant poloidal component of magnetic field in Boozer coordinates "
    + "(proportional to toroidal current)",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["B_theta"],
)
def _I(params, transforms, profiles, data, **kwargs):
    data["I"] = surface_averages(transforms["grid"], data["B_theta"])
    return data


@register_compute_fun(
    name="I_r",
    label="\\partial_{\\rho} I",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Covariant poloidal component of magnetic field in Boozer coordinates "
    + "(proportional to toroidal current), derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["B_theta_r"],
)
def _I_r(params, transforms, profiles, data, **kwargs):
    data["I_r"] = surface_averages(transforms["grid"], data["B_theta_r"])
    return data


@register_compute_fun(
    name="I_rr",
    label="\\partial_{\\rho\\rho} I",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Boozer toroidal current enclosed by flux surfaces, second derivative "
    + "wrt radial coordinate",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["B_theta_rr"],
)
def _I_rr(params, transforms, profiles, data, **kwargs):
    data["I_rr"] = surface_averages(transforms["grid"], data["B_theta_rr"])
    return data


@register_compute_fun(
    name="G",
    label="G",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Covariant toroidal component of magnetic field in Boozer coordinates "
    + "(proportional to poloidal current)",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["B_zeta"],
)
def _G(params, transforms, profiles, data, **kwargs):
    data["G"] = surface_averages(transforms["grid"], data["B_zeta"])
    return data


@register_compute_fun(
    name="G_r",
    label="\\partial_{\\rho} G",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Covariant toroidal component of magnetic field in Boozer coordinates "
    + "(proportional to poloidal current), derivative wrt radial coordinate",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["B_zeta_r"],
)
def _G_r(params, transforms, profiles, data, **kwargs):
    data["G_r"] = surface_averages(transforms["grid"], data["B_zeta_r"])
    return data


@register_compute_fun(
    name="G_rr",
    label="\\partial_{\\rho\\rho} G",
    units="T \\cdot m",
    units_long="Tesla * meters",
    description="Boozer poloidal current enclosed by flux surfaces, second derivative "
    + "wrt radial coordinate",
    dim=1,
    params=[],
    transforms={"grid": []},
    profiles=[],
    coordinates="r",
    data=["B_zeta_rr"],
)
def _G_rr(params, transforms, profiles, data, **kwargs):
    data["G_rr"] = surface_averages(transforms["grid"], data["B_zeta_rr"])
    return data


@register_compute_fun(
    name="current",
    label="\\frac{2\\pi}{\\mu_0} I",
    units="A",
    units_long="Amperes",
    description="Net toroidal current enclosed by flux surfaces",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["I"],
)
def _current(params, transforms, profiles, data, **kwargs):
    data["current"] = 2 * jnp.pi / mu_0 * data["I"]
    return data


@register_compute_fun(
    name="current_r",
    label="\\frac{2\\pi}{\\mu_0} \\partial_{\\rho} I",
    units="A",
    units_long="Amperes",
    description="Net toroidal current enclosed by flux surfaces, derivative "
    + "wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["I_r"],
)
def _current_r(params, transforms, profiles, data, **kwargs):
    data["current_r"] = 2 * jnp.pi / mu_0 * data["I_r"]
    return data


@register_compute_fun(
    name="current_rr",
    label="\\frac{2\\pi}{\\mu_0} \\partial_{\\rho\\rho} I",
    units="A",
    units_long="Amperes",
    description="Net toroidal current enclosed by flux surfaces, second derivative "
    + "wrt radial coordinate",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["I_rr"],
)
def _current_rr(params, transforms, profiles, data, **kwargs):
    data["current_rr"] = 2 * jnp.pi / mu_0 * data["I_rr"]
    return data


@register_compute_fun(
    name="shear",
    label="-\\rho \\frac{\\partial_{\\rho}\\iota}{\\iota}",
    units="~",
    units_long="None",
    description="Global magnetic shear",
    dim=1,
    params=[],
    transforms={},
    profiles=[],
    coordinates="r",
    data=["0", "iota", "iota_r", "rho"],
)
def _shear(params, transforms, profiles, data, **kwargs):
    """Global magnetic shear, as defined in the tokamak literature: -dι/dρ * (ρ/ι).

    When ι=0 ∀ρ (such as in a vacuum tokamak) the shear is defined to be 0 everywhere.
    This implementation is undefined whenever ι=0 otherwise, which is the correct value
    in the case of a Reverse Field Configuration.
    The case where both ι=0 and dι/dρ=0 is rare, and that limit is not implemented.
    """
    eps = 1e2 * jnp.finfo(data["iota"].dtype).eps
    data["shear"] = cond(
        jnp.all(jnp.abs(data["iota"]) < eps),
        lambda _: data["0"],  # if iota is 0 everywhere, set shear to 0 everywhere
        lambda _: -data["rho"] * data["iota_r"] / data["iota"],
        None,
    )
    return data



"""Functions for computing initial guesses for coordinate surfaces."""

import os

import numpy as np

from desc.backend import fori_loop, jit, jnp, put
from desc.basis import zernike_radial
from desc.geometry import FourierRZCurve, Surface
from desc.grid import Grid, _Grid
from desc.io import load
from desc.transform import Transform
from desc.utils import copy_coeffs


def set_initial_guess(eq, *args):  # noqa: C901 - FIXME: simplify this
    """Set the initial guess for the flux surfaces, eg R_lmn, Z_lmn, L_lmn.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to initialize
    args :
        either:
          - No arguments, in which case eq.surface will be scaled for the guess.
          - Another Surface object, which will be scaled to generate the guess.
            Optionally a Curve object may also be supplied for the magnetic axis.
          - Another Equilibrium, whose flux surfaces will be used.
          - File path to a VMEC or DESC equilibrium, which will be loaded and used.
          - Grid and 2-3 ndarrays, specifying the flux surface locations (R, Z, and
            optionally lambda) at fixed flux coordinates. All arrays should have the
            same length. Optionally, an ndarray of shape(k,3) may be passed instead
            of a grid.

    Examples
    --------
    Use existing equil.surface and scales down for guess:

    >>> equil.set_initial_guess()

    Use supplied Surface and scales down for guess. Assumes axis is centroid
    of user supplied surface:

    >>> equil.set_initial_guess(surface)

    Optionally, an interior surface may be scaled by giving the surface a
    flux label:

    >>> surf = FourierRZToroidalSurface(rho=0.7)
    >>> equil.set_initial_guess(surf)

    Use supplied Surface and a supplied Curve for axis and scales between
    them for guess:

    >>> equil.set_initial_guess(surface, curve)

    Use the flux surfaces from an existing Equilibrium:

    >>> equil.set_initial_guess(equil2)

    Use flux surfaces from existing Equilibrium or VMEC output stored on disk:

    >>> equil.set_initial_guess(path_to_saved_DESC_or_VMEC_output)

    Use flux surfaces specified by points:
    nodes should either be a Grid or an ndarray, shape(k,3) giving the locations
    in rho, theta, zeta coordinates. R, Z, and optionally lambda should be
    array-like, shape(k,) giving the corresponding real space coordinates

    >>> equil.set_initial_guess(nodes, R, Z, lambda)

    """
    nargs = len(args)
    if nargs > 4:
        raise ValueError(
            "set_initial_guess should be called with 4 or fewer arguments."
        )
    if nargs == 0 or nargs == 1 and args[0] is None:
        if hasattr(eq, "_surface"):
            # use whatever surface is already assigned
            if hasattr(eq, "_axis"):
                axisR = np.array([eq._axis.R_basis.modes[:, -1], eq._axis.R_n]).T
                axisZ = np.array([eq._axis.Z_basis.modes[:, -1], eq._axis.Z_n]).T
            else:
                axisR = None
                axisZ = None
            coord = eq.surface.rho if hasattr(eq.surface, "rho") else None
            eq.R_lmn = _initial_guess_surface(
                eq.R_basis,
                eq.Rb_lmn,
                eq.surface.R_basis,
                axisR,
                coord=coord,
            )
            eq.Z_lmn = _initial_guess_surface(
                eq.Z_basis,
                eq.Zb_lmn,
                eq.surface.Z_basis,
                axisZ,
                coord=coord,
            )
        else:
            raise ValueError(
                "set_initial_guess called with no arguments, "
                + "but no surface is assigned."
            )
    else:  # nargs > 0
        if isinstance(args[0], Surface):
            surface = args[0]
            if nargs > 1:
                if isinstance(args[1], FourierRZCurve):
                    axis = args[1]
                    axisR = np.array([axis.R_basis.modes[:, -1], axis.R_n]).T
                    axisZ = np.array([axis.Z_basis.modes[:, -1], axis.Z_n]).T
                else:
                    raise TypeError(
                        "Don't know how to initialize from object type {}".format(
                            type(args[1])
                        )
                    )
            else:
                axisR = None
                axisZ = None
            coord = surface.rho if hasattr(surface, "rho") else None
            eq.R_lmn = _initial_guess_surface(
                eq.R_basis,
                surface.R_lmn,
                surface.R_basis,
                axisR,
                coord=coord,
            )
            eq.Z_lmn = _initial_guess_surface(
                eq.Z_basis,
                surface.Z_lmn,
                surface.Z_basis,
                axisZ,
                coord=coord,
            )
        elif type(args[0]) is type(eq):
            eq1 = args[0]
            if nargs > 1:
                raise ValueError(
                    "set_initial_guess got unknown additional argument {}.".format(
                        args[1]
                    )
                )
            eq.R_lmn = copy_coeffs(eq1.R_lmn, eq1.R_basis.modes, eq.R_basis.modes)
            eq.Z_lmn = copy_coeffs(eq1.Z_lmn, eq1.Z_basis.modes, eq.Z_basis.modes)
            eq.L_lmn = copy_coeffs(eq1.L_lmn, eq1.L_basis.modes, eq.L_basis.modes)
            eq.Ra_n = copy_coeffs(
                eq1.Ra_n, eq1.axis.R_basis.modes, eq.axis.R_basis.modes
            )
            eq.Za_n = copy_coeffs(
                eq1.Za_n, eq1.axis.Z_basis.modes, eq.axis.Z_basis.modes
            )

        elif isinstance(args[0], (str, os.PathLike)):
            # from file
            path = args[0]
            file_format = None
            if nargs > 1:
                if isinstance(args[1], str):
                    file_format = args[1]
                else:
                    raise ValueError(
                        "set_initial_guess got unknown additional argument "
                        + "{}.".format(args[1])
                    )
            try:  # is it desc?
                eq1 = load(path, file_format)
            except:  # noqa: E722
                try:  # maybe its vmec
                    from desc.vmec import VMECIO

                    eq1 = VMECIO.load(path)
                except:  # noqa: E722
                    raise ValueError(
                        "Could not load equilibrium from path {}, ".format(path)
                        + "please make sure it is a valid DESC or VMEC equilibrium."
                    )
            if not type(eq1) is type(eq):
                if hasattr(eq1, "equilibria"):  # it's a family!
                    eq1 = eq1[-1]
                else:
                    raise TypeError(
                        "Cannot initialize equilibrium from loaded object of type "
                        + "{}".format(type(eq1))
                    )
            eq.R_lmn = copy_coeffs(eq1.R_lmn, eq1.R_basis.modes, eq.R_basis.modes)
            eq.Z_lmn = copy_coeffs(eq1.Z_lmn, eq1.Z_basis.modes, eq.Z_basis.modes)
            eq.L_lmn = copy_coeffs(eq1.L_lmn, eq1.L_basis.modes, eq.L_basis.modes)
            eq.Ra_n = copy_coeffs(
                eq1.Ra_n, eq1.axis.R_basis.modes, eq.axis.R_basis.modes
            )
            eq.Za_n = copy_coeffs(
                eq1.Za_n, eq1.axis.Z_basis.modes, eq.axis.Z_basis.modes
            )

        elif nargs > 2:  # assume we got nodes and ndarray of points
            grid = args[0]
            R = args[1]
            eq.R_lmn = _initial_guess_points(grid, R, eq.R_basis)
            Z = args[2]
            eq.Z_lmn = _initial_guess_points(grid, Z, eq.Z_basis)
            if nargs > 3:
                lmbda = args[3]
                eq.L_lmn = _initial_guess_points(grid, lmbda, eq.L_basis)
            else:
                eq.L_lmn = jnp.zeros(eq.L_basis.num_modes)

        else:
            raise ValueError("Can't initialize equilibrium from args {}.".format(args))
    return eq


def _initial_guess_surface(x_basis, b_lmn, b_basis, axis=None, mode=None, coord=None):
    """Create an initial guess from boundary coefficients and a magnetic axis guess.

    Parameters
    ----------
    x_basis : FourierZernikeBasis
        basis of the flux surfaces (for R, Z, or Lambda).
    b_lmn : ndarray, shape(b_basis.num_modes,)
        vector of boundary coefficients associated with b_basis.
    b_basis : Basis
        basis of the boundary surface (for Rb or Zb)
    axis : ndarray, shape(num_modes,2)
        coefficients of the magnetic axis. axis[i, :] = [n, x0].
        Only used for 'lcfs' boundary mode. Defaults to m=0 modes of boundary
    mode : str
        One of 'lcfs', 'poincare'.
        Whether the boundary condition is specified by the last closed flux surface
        (rho=1) or the Poincare section (zeta=0).
    coord : float or None
        Surface label (ie, rho, zeta, etc.) for supplied surface.

    Returns
    -------
    x_lmn : ndarray
        vector of flux surface coefficients associated with x_basis.

    """
    b_modes = jnp.asarray(b_basis.modes)
    x_modes = jnp.asarray(x_basis.modes)
    b_lmn = jnp.asarray(b_lmn)
    x_lmn = jnp.zeros((x_basis.num_modes,))

    if mode is None:
        # auto-detect based on mode numbers
        if np.all(b_basis.modes[:, 0] == 0):
            mode = "lcfs"
        elif np.all(b_basis.modes[:, 2] == 0):
            mode = "poincare"
        else:
            raise ValueError("Surface should have either l=0 or n=0")
    if mode == "lcfs":
        if coord is None:
            coord = 1.0
        if axis is None:
            axidx = np.where(b_basis.modes[:, 1] == 0)[0]
            axis = jnp.array([b_basis.modes[axidx, 2], b_lmn[axidx]]).T

        # first do all the m != 0 modes, easiest since no special logic needed
        def body(k, x_lmn):
            l, m, n = b_modes[k]
            scale = zernike_radial(coord, abs(m), m)
            # index of basis mode with lowest radial power (l = |m|)
            mask0 = (x_modes == jnp.array([abs(m), m, n])).all(axis=1)
            x_lmn = jnp.where(mask0, b_lmn[k] / scale, x_lmn)
            return x_lmn

        x_lmn = fori_loop(0, b_basis.num_modes, body, x_lmn)

        # now overwrite stuff to deal with the axis
        scale = zernike_radial(coord, 0, 0)
        for k, (l, m, n) in enumerate(b_basis.modes):
            if m != 0:
                continue
            # index of basis mode with lowest radial power (l = |m|)
            idx0 = np.where((x_basis.modes == [abs(m), m, n]).all(axis=1))[0]
            # index of basis mode with second lowest radial power (l = |m| + 2)
            idx2 = np.where((x_basis.modes == [abs(m) + 2, m, n]).all(axis=1))[0]
            ax = np.where(axis[:, 0] == n)[0]
            if ax.size:
                a_n = axis[ax[0], 1]  # use provided axis guess
            else:
                a_n = b_lmn[k]  # use boundary centroid as axis
            x_lmn = jit(put)(x_lmn, idx0, (b_lmn[k] + a_n) / 2 / scale)
            x_lmn = jit(put)(x_lmn, idx2, (b_lmn[k] - a_n) / 2 / scale)

    elif mode == "poincare":

        def body(k, x_lmn):
            l, m, n = b_modes[k]
            mask0 = (x_modes == jnp.array([l, m, n])).all(axis=1)
            x_lmn = jnp.where(mask0, b_lmn[k], x_lmn)
            return x_lmn

        x_lmn = fori_loop(0, b_basis.num_modes, body, x_lmn)

    else:
        raise ValueError("Boundary mode should be either 'lcfs' or 'poincare'.")

    return x_lmn


def _initial_guess_points(nodes, x, x_basis):
    """Create an initial guess based on locations of flux surfaces in real space.

    Parameters
    ----------
    nodes : Grid or ndarray, shape(k,3)
        Locations in flux coordinates where real space coordinates are given.
    x : ndarray, shape(k,)
        R, Z or lambda values at specified nodes.
    x_basis : Basis
        Spectral basis for x (R, Z or lambda)

    Returns
    -------
    x_lmn : ndarray
        Vector of flux surface coefficients associated with x_basis.

    """
    if not isinstance(nodes, _Grid):
        nodes = Grid(nodes, sort=False)
    transform = Transform(nodes, x_basis, build=False, build_pinv=True)
    x_lmn = transform.fit(x)
    return x_lmn



"""Regression tests to verify that DESC agrees with VMEC and itself.

Computes several benchmark equilibria and compares the solutions by measuring the
difference in areas between constant theta and rho contours.
"""

import numpy as np
import pytest
from qic import Qic
from qsc import Qsc

import desc.examples
from desc.continuation import _solve_axisym, solve_continuation_automatic
from desc.equilibrium import EquilibriaFamily, Equilibrium
from desc.geometry import FourierRZToroidalSurface
from desc.grid import LinearGrid
from desc.io import load
from desc.objectives import (
    AspectRatio,
    CurrentDensity,
    FixBoundaryR,
    FixBoundaryZ,
    FixCurrent,
    FixIota,
    FixPressure,
    FixPsi,
    FixSumModesLambda,
    ForceBalance,
    ForceBalanceAnisotropic,
    HelicalForceBalance,
    ObjectiveFunction,
    QuasisymmetryBoozer,
    QuasisymmetryTwoTerm,
    RadialForceBalance,
    get_fixed_boundary_constraints,
    get_NAE_constraints,
)
from desc.optimize import Optimizer
from desc.profiles import FourierZernikeProfile, PowerSeriesProfile
from desc.vmec_utils import vmec_boundary_subspace

from .utils import area_difference_desc, area_difference_vmec


@pytest.mark.regression
@pytest.mark.solve
def test_SOLOVEV_vacuum(SOLOVEV_vac):
    """Tests that the SOLOVEV vacuum example gives no rotational transform."""
    eq = EquilibriaFamily.load(load_from=str(SOLOVEV_vac["desc_h5_path"]))[-1]
    data = eq.compute("|J|")

    np.testing.assert_allclose(data["iota"], 0, atol=1e-16)
    np.testing.assert_allclose(data["|J|"], 0, atol=3e-3)

    # test that solving with the continuation method works correctly
    # when eq resolution is lower than the mres_step
    eq.change_resolution(L=3, M=3)
    eqf = _solve_axisym(eq, mres_step=6)
    assert len(eqf) == 1
    assert eqf[-1].L == eq.L
    assert eqf[-1].M == eq.M
    assert eqf[-1].N == eq.N


@pytest.mark.regression
@pytest.mark.solve
def test_SOLOVEV_results(SOLOVEV):
    """Tests that the SOLOVEV example gives the same result as VMEC."""
    eq = EquilibriaFamily.load(load_from=str(SOLOVEV["desc_h5_path"]))[-1]
    rho_err, theta_err = area_difference_vmec(eq, SOLOVEV["vmec_nc_path"])

    np.testing.assert_allclose(rho_err, 0, atol=1e-3)
    np.testing.assert_allclose(theta_err, 0, atol=1e-4)


@pytest.mark.regression
@pytest.mark.solve
def test_SOLOVEV_anisotropic_results(SOLOVEV):
    """Tests that SOLOVEV with zero anisotropic pressure gives the same result."""
    eq = EquilibriaFamily.load(load_from=str(SOLOVEV["desc_h5_path"]))[-1]
    # reset to start
    eq.set_initial_guess()
    # give it a zero anisotropy profile
    anisotropy = FourierZernikeProfile()
    anisotropy.change_resolution(eq.L, eq.M, eq.N)
    eq.anisotropy = anisotropy

    obj = ObjectiveFunction(ForceBalanceAnisotropic(eq=eq))
    constraints = get_fixed_boundary_constraints(eq=eq, anisotropy=True)
    eq.solve(obj, constraints, verbose=3)
    rho_err, theta_err = area_difference_vmec(eq, SOLOVEV["vmec_nc_path"])

    np.testing.assert_allclose(rho_err, 0, atol=1e-3)
    np.testing.assert_allclose(theta_err, 0, atol=1e-4)


@pytest.mark.unit
@pytest.mark.solve
def test_DSHAPE_results(DSHAPE):
    """Tests that the DSHAPE examples gives the same results as VMEC."""
    eq = EquilibriaFamily.load(load_from=str(DSHAPE["desc_h5_path"]))[-1]
    rho_err, theta_err = area_difference_vmec(eq, DSHAPE["vmec_nc_path"])
    np.testing.assert_allclose(rho_err, 0, atol=2e-3)
    np.testing.assert_allclose(theta_err, 0, atol=1e-4)


@pytest.mark.unit
@pytest.mark.solve
def test_DSHAPE_current_results(DSHAPE_current):
    """Tests that the DSHAPE with fixed current gives the same results as VMEC."""
    eq = EquilibriaFamily.load(load_from=str(DSHAPE_current["desc_h5_path"]))[-1]
    rho_err, theta_err = area_difference_vmec(eq, DSHAPE_current["vmec_nc_path"])
    np.testing.assert_allclose(rho_err, 0, atol=3e-3)
    np.testing.assert_allclose(theta_err, 0, atol=1e-4)


@pytest.mark.regression
@pytest.mark.solve
def test_HELIOTRON_results(HELIOTRON):
    """Tests that the HELIOTRON examples gives the same results as VMEC."""
    eq = EquilibriaFamily.load(load_from=str(HELIOTRON["desc_h5_path"]))[-1]
    rho_err, theta_err = area_difference_vmec(eq, HELIOTRON["vmec_nc_path"])
    np.testing.assert_allclose(rho_err.mean(), 0, atol=2e-2)
    np.testing.assert_allclose(theta_err.mean(), 0, atol=2e-2)


@pytest.mark.unit
@pytest.mark.solve
def test_HELIOTRON_vac_results(HELIOTRON_vac):
    """Tests that the HELIOTRON examples gives the same results as VMEC."""
    eq = EquilibriaFamily.load(load_from=str(HELIOTRON_vac["desc_h5_path"]))[-1]
    rho_err, theta_err = area_difference_vmec(eq, HELIOTRON_vac["vmec_nc_path"])
    np.testing.assert_allclose(rho_err.mean(), 0, atol=1e-2)
    np.testing.assert_allclose(theta_err.mean(), 0, atol=2e-2)
    curr = eq.get_profile("current")
    np.testing.assert_allclose(curr(np.linspace(0, 1, 20)), 0, atol=1e-8)


@pytest.mark.regression
@pytest.mark.solve
def test_HELIOTRON_vac2_results(HELIOTRON_vac, HELIOTRON_vac2):
    """Tests that the 2 methods for solving vacuum give the same results."""
    eq1 = EquilibriaFamily.load(load_from=str(HELIOTRON_vac["desc_h5_path"]))[-1]
    eq2 = EquilibriaFamily.load(load_from=str(HELIOTRON_vac2["desc_h5_path"]))[-1]
    rho_err, theta_err = area_difference_desc(eq1, eq2)
    np.testing.assert_allclose(rho_err[:, 4:], 0, atol=1e-2)
    np.testing.assert_allclose(theta_err, 0, atol=1e-4)
    curr1 = eq1.get_profile("current")
    curr2 = eq2.get_profile("current")
    iota1 = eq1.get_profile("iota")
    iota2 = eq2.get_profile("iota")
    np.testing.assert_allclose(curr1(np.linspace(0, 1, 20)), 0, atol=1e-8)
    np.testing.assert_allclose(curr2(np.linspace(0, 1, 20)), 0, atol=1e-8)
    np.testing.assert_allclose(iota1.params, iota2.params, rtol=1e-1, atol=1e-1)


@pytest.mark.regression
def test_force_balance_grids():
    """Compares radial & helical force balance on same vs different grids."""
    # When ConcentricGrid had a rotation option, RadialForceBalance, HelicalForceBalance
    # defaulted to cos, sin rotation, respectively.
    # This test has been kept to increase code coverage.

    def test(iota=False):
        if iota:
            eq1 = Equilibrium(iota=PowerSeriesProfile(0), sym=True)
            eq2 = Equilibrium(iota=PowerSeriesProfile(0), sym=True)
        else:
            eq1 = Equilibrium(current=PowerSeriesProfile(0), sym=True)
            eq2 = Equilibrium(current=PowerSeriesProfile(0), sym=True)

        res = 3
        eq1.change_resolution(L=res, M=res)
        eq1.L_grid = res
        eq1.M_grid = res
        eq2.change_resolution(L=res, M=res)
        eq2.L_grid = res
        eq2.M_grid = res

        # force balances on the same grids
        obj1 = ObjectiveFunction(ForceBalance(eq=eq1))
        eq1.solve(objective=obj1)

        # force balances on different grids
        obj2 = ObjectiveFunction(
            (RadialForceBalance(eq=eq2), HelicalForceBalance(eq=eq2))
        )
        eq2.solve(objective=obj2)

        np.testing.assert_allclose(eq1.R_lmn, eq2.R_lmn, atol=5e-4)
        np.testing.assert_allclose(eq1.Z_lmn, eq2.Z_lmn, atol=5e-4)
        np.testing.assert_allclose(eq1.L_lmn, eq2.L_lmn, atol=2e-3)

    test(iota=True)
    test(iota=False)


@pytest.mark.regression
def test_solve_bounds():
    """Tests optimizing with bounds=(lower bound, upper bound)."""
    # decrease resolution and double pressure so no longer in force balance
    eq = desc.examples.get("DSHAPE")
    eq.change_resolution(L=eq.M, L_grid=eq.M_grid)
    eq.p_l *= 2

    # target force balance residuals with |F| <= 1e3 N
    obj = ObjectiveFunction(
        ForceBalance(normalize=False, normalize_target=False, bounds=(-1e3, 1e3), eq=eq)
    )
    eq.solve(objective=obj, ftol=1e-16, xtol=1e-16, maxiter=200, verbose=3)

    # check that all errors are nearly 0, since residual values are within target bounds
    f = obj.compute_scaled_error(obj.x(eq))
    np.testing.assert_allclose(f, 0, atol=1e-4)


@pytest.mark.regression
def test_1d_optimization():
    """Tests 1D optimization for target aspect ratio."""
    eq = desc.examples.get("SOLOVEV")
    objective = ObjectiveFunction(AspectRatio(eq=eq, target=2.5))
    constraints = (
        ForceBalance(eq=eq),
        FixBoundaryR(eq=eq),
        FixBoundaryZ(eq=eq, modes=eq.surface.Z_basis.modes[0:-1, :]),
        FixPressure(eq=eq),
        FixIota(eq=eq),
        FixPsi(eq=eq),
    )
    options = {"perturb_options": {"order": 1}}
    with pytest.warns((FutureWarning, UserWarning)):
        eq.optimize(objective, constraints, optimizer="lsq-exact", options=options)

    np.testing.assert_allclose(eq.compute("R0/a")["R0/a"], 2.5, rtol=2e-4)


@pytest.mark.regression
def test_1d_optimization_old():
    """Tests 1D optimization for target aspect ratio."""
    eq = desc.examples.get("SOLOVEV")
    objective = ObjectiveFunction(AspectRatio(eq=eq, target=2.5))
    eq._optimize(
        objective,
        copy=False,
        solve_options={"verbose": 0},
        perturb_options={
            "dZb": True,
            "subspace": vmec_boundary_subspace(eq, ZBS=[0, 1]),
        },
    )

    np.testing.assert_allclose(eq.compute("R0/a")["R0/a"], 2.5, rtol=1e-3)


def run_qh_step(n, eq):
    """Run 1 step of the precise QH optimization example from Landreman & Paul."""
    print(f"==========QH step {n+1}==========")
    grid = LinearGrid(
        M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, rho=np.array([0.6, 0.8, 1.0]), sym=True
    )

    objective = ObjectiveFunction(
        (
            QuasisymmetryTwoTerm(eq=eq, helicity=(1, eq.NFP), grid=grid),
            AspectRatio(eq=eq, target=8, weight=1e2),
        ),
        verbose=0,
    )
    R_modes = np.vstack(
        (
            [0, 0, 0],
            eq.surface.R_basis.modes[
                np.max(np.abs(eq.surface.R_basis.modes), 1) > n + 1, :
            ],
        )
    )
    Z_modes = eq.surface.Z_basis.modes[
        np.max(np.abs(eq.surface.Z_basis.modes), 1) > n + 1, :
    ]
    constraints = (
        ForceBalance(eq=eq),
        FixBoundaryR(eq=eq, modes=R_modes),
        FixBoundaryZ(eq=eq, modes=Z_modes),
        FixPressure(eq=eq),
        FixCurrent(eq=eq),
        FixPsi(eq=eq),
    )
    optimizer = Optimizer("proximal-lsq-exact")
    eq1, history = eq.optimize(
        objective=objective,
        constraints=constraints,
        optimizer=optimizer,
        maxiter=50,
        verbose=3,
        copy=True,
        options={
            "initial_trust_ratio": 1.0,  # for backwards consistency
            "perturb_options": {"verbose": 0},
            "solve_options": {"verbose": 0},
        },
    )

    return eq1


@pytest.mark.regression
@pytest.mark.slow
def test_qh_optimization():
    """Tests first 3 steps of precise QH optimization."""
    # create initial equilibrium
    surf = FourierRZToroidalSurface(
        R_lmn=[1, 0.125, 0.1],
        Z_lmn=[-0.125, -0.1],
        modes_R=[[0, 0], [1, 0], [0, 1]],
        modes_Z=[[-1, 0], [0, -1]],
        NFP=4,
    )
    eq = Equilibrium(M=4, N=4, Psi=0.04, surface=surf)
    eq = solve_continuation_automatic(eq, objective="force", bdry_step=0.5, verbose=3)[
        -1
    ]

    eq1 = run_qh_step(0, eq)

    obj = QuasisymmetryBoozer(helicity=(1, eq1.NFP), eq=eq1)
    obj.build()
    B_asym = obj.compute(*obj.xs(eq1))

    np.testing.assert_array_less(np.abs(B_asym).max(), 1e-1)
    np.testing.assert_array_less(eq1.compute("a_major/a_minor")["a_major/a_minor"], 5)

    eq2 = run_qh_step(1, eq1)

    obj = QuasisymmetryBoozer(helicity=(1, eq2.NFP), eq=eq2)
    obj.build()
    B_asym = obj.compute(*obj.xs(eq2))
    np.testing.assert_array_less(np.abs(B_asym).max(), 1e-2)
    np.testing.assert_array_less(eq2.compute("a_major/a_minor")["a_major/a_minor"], 5)

    eq3 = run_qh_step(2, eq2)

    obj = QuasisymmetryBoozer(helicity=(1, eq3.NFP), eq=eq3)
    obj.build()
    B_asym = obj.compute(*obj.xs(eq3))
    np.testing.assert_array_less(np.abs(B_asym).max(), 2e-3)
    np.testing.assert_array_less(eq3.compute("a_major/a_minor")["a_major/a_minor"], 5)


@pytest.mark.regression
@pytest.mark.solve
def test_ATF_results(tmpdir_factory):
    """Test automatic continuation method with ATF."""
    output_dir = tmpdir_factory.mktemp("result")
    eq0 = desc.examples.get("ATF")
    eq = Equilibrium(
        Psi=eq0.Psi,
        NFP=eq0.NFP,
        L=eq0.L,
        M=eq0.M,
        N=eq0.N,
        L_grid=eq0.L_grid,
        M_grid=eq0.M_grid,
        N_grid=eq0.N_grid,
        pressure=eq0.pressure,
        iota=eq0.iota,
        surface=eq0.get_surface_at(rho=1),
        sym=eq0.sym,
        spectral_indexing=eq0.spectral_indexing,
    )
    eqf = EquilibriaFamily.solve_continuation_automatic(
        eq,
        verbose=2,
        checkpoint_path=output_dir.join("ATF.h5"),
    )
    eqf = load(output_dir.join("ATF.h5"))
    rho_err, theta_err = area_difference_desc(eq0, eqf[-1])
    np.testing.assert_allclose(rho_err[:, 4:], 0, atol=4e-2)
    np.testing.assert_allclose(theta_err, 0, atol=5e-4)


@pytest.mark.regression
@pytest.mark.solve
def test_ESTELL_results(tmpdir_factory):
    """Test automatic continuation method with ESTELL."""
    output_dir = tmpdir_factory.mktemp("result")
    eq0 = desc.examples.get("ESTELL")
    eq = Equilibrium(
        Psi=eq0.Psi,
        NFP=eq0.NFP,
        L=eq0.L,
        M=eq0.M,
        N=eq0.N,
        L_grid=eq0.L_grid,
        M_grid=eq0.M_grid,
        N_grid=eq0.N_grid,
        pressure=eq0.pressure,
        current=eq0.current,
        surface=eq0.get_surface_at(rho=1),
        sym=eq0.sym,
        spectral_indexing=eq0.spectral_indexing,
    )
    eqf = EquilibriaFamily.solve_continuation_automatic(
        eq,
        verbose=2,
        checkpoint_path=output_dir.join("ESTELL.h5"),
    )
    eqf = load(output_dir.join("ESTELL.h5"))
    rho_err, theta_err = area_difference_desc(eq0, eqf[-1])
    np.testing.assert_allclose(rho_err[:, 4:], 0, atol=5e-2)
    np.testing.assert_allclose(theta_err, 0, atol=1e-4)


@pytest.mark.regression
@pytest.mark.solve
def test_simsopt_QH_comparison():
    """Test case that previously stalled before getting to the solution.

    From Matt's comparison with SIMSOPT.
    """
    nfp = 4
    aspect_target = 8.0
    # Initial (m=0, n=nfp) mode of the axis:
    torsion = 0.4
    LMN_resolution = 6
    # Set shape of the initial condition.
    # R_lmn and Z_lmn are the amplitudes. modes_R and modes_Z are the (m,n) pairs.
    surface = FourierRZToroidalSurface(
        R_lmn=[1.0, 1.0 / aspect_target, torsion],
        modes_R=[[0, 0], [1, 0], [0, 1]],
        Z_lmn=[0, -1.0 / aspect_target, torsion],
        modes_Z=[[0, 0], [-1, 0], [0, -1]],
        NFP=nfp,
    )
    # Set up a vacuum field:
    pressure = PowerSeriesProfile(params=[0], modes=[0])
    ndofs_current = 3
    current = PowerSeriesProfile(
        params=np.zeros(ndofs_current),
        modes=2 * np.arange(ndofs_current),
    )
    eq = Equilibrium(
        surface=surface,
        pressure=pressure,
        current=current,
        Psi=np.pi / (aspect_target**2),  # So |B| is ~ 1 T.
        NFP=nfp,
        L=LMN_resolution,
        M=LMN_resolution,
        N=LMN_resolution,
        L_grid=2 * LMN_resolution,
        M_grid=2 * LMN_resolution,
        N_grid=2 * LMN_resolution,
        sym=True,
    )
    # Fix the major radius, and all modes with |m| or |n| > 1:
    R_modes_to_fix = []
    for j in range(eq.surface.R_basis.modes.shape[0]):
        m = eq.surface.R_basis.modes[j, 1]
        n = eq.surface.R_basis.modes[j, 2]
        if (m * n < 0) or (m == 0 and n == 0) or (abs(m) > 1) or (abs(n) > 1):
            R_modes_to_fix.append([0, m, n])
        else:
            print(f"Freeing R mode: m={m}, n={n}")
    R_modes_to_fix = np.array(R_modes_to_fix)
    Z_modes_to_fix = []
    for j in range(eq.surface.Z_basis.modes.shape[0]):
        m = eq.surface.Z_basis.modes[j, 1]
        n = eq.surface.Z_basis.modes[j, 2]
        if (m * n > 0) or (m == 0 and n == 0) or (abs(m) > 1) or (abs(n) > 1):
            Z_modes_to_fix.append([0, m, n])
        else:
            print(f"Freeing Z mode: m={m}, n={n}")
    Z_modes_to_fix = np.array(Z_modes_to_fix)
    eq.solve(
        verbose=3,
        ftol=1e-8,
        constraints=get_fixed_boundary_constraints(eq=eq, profiles=False),
        objective=ObjectiveFunction(objectives=CurrentDensity(eq=eq)),
    )
    ##################################
    # Done creating initial condition.
    # Now define optimization problem.
    ##################################
    constraints = (
        CurrentDensity(eq=eq),
        FixBoundaryR(eq=eq, modes=R_modes_to_fix),
        FixBoundaryZ(eq=eq, modes=Z_modes_to_fix),
        FixPressure(eq=eq),
        FixCurrent(eq=eq),
        FixPsi(eq=eq),
    )
    # Objective function, for both desc and simsopt:
    # f = (aspect - 8)^2 + (2pi)^{-2} \int dtheta \int d\zeta [f_C(rho=1)]^2
    # grid for quasisymmetry objective:
    grid = LinearGrid(
        M=eq.M,
        N=eq.N,
        rho=[1.0],
        NFP=nfp,
    )
    # Cancel factor of 1/2 in desc objective which is not present in simsopt:
    aspect_weight = np.sqrt(2)
    # Also scale QS weight to match simsopt/VMEC
    qs_weight = np.sqrt(1 / (8 * (np.pi**4)))
    objective = ObjectiveFunction(
        (
            AspectRatio(
                eq=eq, target=aspect_target, weight=aspect_weight, normalize=False
            ),
            QuasisymmetryTwoTerm(
                eq=eq, helicity=(-1, nfp), grid=grid, weight=qs_weight, normalize=False
            ),
        )
    )
    eq2, result = eq.optimize(
        verbose=3,
        objective=objective,
        constraints=constraints,
        optimizer=Optimizer("proximal-lsq-exact"),
        ftol=1e-3,
    )
    aspect = eq2.compute("R0/a")["R0/a"]
    np.testing.assert_allclose(aspect, aspect_target, atol=1e-2, rtol=1e-2)
    np.testing.assert_array_less(objective.compute_scalar(objective.x(eq)), 0.075)
    np.testing.assert_array_less(eq2.compute("a_major/a_minor")["a_major/a_minor"], 5)


@pytest.mark.regression
@pytest.mark.solve
@pytest.mark.slow
def test_NAE_QSC_solve():
    """Test O(rho) NAE QSC constraints solve."""
    qsc = Qsc.from_paper("precise QA")
    ntheta = 75
    r = 0.01
    N = 9
    eq = Equilibrium.from_near_axis(qsc, r=r, L=6, M=6, N=N, ntheta=ntheta)

    orig_Rax_val = eq.axis.R_n
    orig_Zax_val = eq.axis.Z_n

    eq_fit = eq.copy()
    eq_lambda_fixed_0th_order = eq.copy()
    eq_lambda_fixed_1st_order = eq.copy()

    # this has all the constraints we need,
    #  iota=False specifies we want to fix current instead of iota
    cs = get_NAE_constraints(eq, qsc, iota=False, order=1, fix_lambda=False, N=eq.N)
    cs_lambda_fixed_0th_order = get_NAE_constraints(
        eq_lambda_fixed_0th_order, qsc, iota=False, order=1, fix_lambda=0, N=eq.N
    )
    cs_lambda_fixed_1st_order = get_NAE_constraints(
        eq_lambda_fixed_1st_order, qsc, iota=False, order=1, fix_lambda=True, N=eq.N
    )

    for c in cs:
        # should be no FixSumModeslambda in the fix_lambda=False constraint
        assert not isinstance(c, FixSumModesLambda)

    for eqq, constraints in zip(
        [eq, eq_lambda_fixed_0th_order, eq_lambda_fixed_1st_order],
        [cs, cs_lambda_fixed_0th_order, cs_lambda_fixed_1st_order],
    ):
        objectives = ForceBalance(eq=eqq)
        obj = ObjectiveFunction(objectives)
        print(constraints)

        eqq.solve(
            verbose=3,
            ftol=1e-2,
            objective=obj,
            maxiter=100,
            xtol=1e-6,
            constraints=constraints,
        )
    grid = LinearGrid(L=10, M=20, N=20, NFP=eq.NFP, sym=True, axis=False)
    grid_axis = LinearGrid(rho=0.0, theta=0.0, N=eq.N_grid, NFP=eq.NFP)
    # Make sure axis is same
    for eqq, string in zip(
        [eq, eq_lambda_fixed_0th_order, eq_lambda_fixed_1st_order],
        ["no lambda constraint", "lambda_fixed_0th_order", "lambda_fixed_1st_order"],
    ):
        np.testing.assert_array_almost_equal(orig_Rax_val, eqq.axis.R_n, err_msg=string)
        np.testing.assert_array_almost_equal(orig_Zax_val, eqq.axis.Z_n, err_msg=string)

        # Make sure surfaces of solved equilibrium are similar near axis as QSC
        rho_err, theta_err = area_difference_desc(eqq, eq_fit)

        np.testing.assert_allclose(rho_err[:, 0:-4], 0, atol=1e-2, err_msg=string)
        np.testing.assert_allclose(theta_err[:, 0:-6], 0, atol=1e-3, err_msg=string)

        # Make sure iota of solved equilibrium is same near axis as QSC

        iota = grid.compress(eqq.compute("iota", grid=grid)["iota"])

        np.testing.assert_allclose(iota[0], qsc.iota, atol=1e-5, err_msg=string)
        np.testing.assert_allclose(iota[1:10], qsc.iota, atol=1e-3, err_msg=string)

        ### check lambda to match near axis
        # Evaluate lambda near the axis
        data_nae = eqq.compute(["lambda", "|B|"], grid=grid_axis)
        lam_nae = data_nae["lambda"]
        # Reshape to form grids on theta and phi

        phi = np.squeeze(grid_axis.nodes[:, 2])
        np.testing.assert_allclose(
            lam_nae, -qsc.iota * qsc.nu_spline(phi), atol=2e-5, err_msg=string
        )

        # check |B| on axis
        np.testing.assert_allclose(
            data_nae["|B|"], np.ones(np.size(phi)) * qsc.B0, atol=1e-4, err_msg=string
        )


@pytest.mark.regression
@pytest.mark.solve
@pytest.mark.slow
def test_NAE_QIC_solve():
    """Test O(rho) NAE QIC constraints solve."""
    # get Qic example
    qic = Qic.from_paper("QI NFP2 r2", nphi=301, order="r1")
    qic.lasym = False  # don't need to consider stellarator asym for order 1 constraints
    ntheta = 75
    r = 0.01
    N = 11
    eq = Equilibrium.from_near_axis(qic, r=r, L=7, M=7, N=N, ntheta=ntheta)

    orig_Rax_val = eq.axis.R_n
    orig_Zax_val = eq.axis.Z_n

    eq_fit = eq.copy()

    # this has all the constraints we need,
    #  iota=False specifies we want to fix current instead of iota
    cs = get_NAE_constraints(eq, qic, iota=False, order=1)

    objectives = ForceBalance(eq=eq)
    obj = ObjectiveFunction(objectives)

    eq.solve(
        verbose=3, ftol=1e-2, objective=obj, maxiter=100, xtol=1e-6, constraints=cs
    )

    # Make sure axis is same
    np.testing.assert_array_almost_equal(orig_Rax_val, eq.axis.R_n)
    np.testing.assert_array_almost_equal(orig_Zax_val, eq.axis.Z_n)

    # Make sure surfaces of solved equilibrium are similar near axis as QIC
    rho_err, theta_err = area_difference_desc(eq, eq_fit)

    np.testing.assert_allclose(rho_err[:, 0:3], 0, atol=5e-2)
    # theta error isn't really an indicator of near axis behavior
    # since it's computed over the full radius, but just indicates that
    # eq is similar to eq_fit
    np.testing.assert_allclose(theta_err, 0, atol=5e-2)

    # Make sure iota of solved equilibrium is same near axis as QIC
    grid = LinearGrid(L=10, M=20, N=20, NFP=eq.NFP, sym=True, axis=False)
    iota = grid.compress(eq.compute("iota", grid=grid)["iota"])

    np.testing.assert_allclose(iota[1], qic.iota, atol=2e-5)
    np.testing.assert_allclose(iota[1:10], qic.iota, atol=5e-4)

    # check lambda to match near axis
    grid_2d_05 = LinearGrid(rho=np.array(1e-6), M=50, N=50, NFP=eq.NFP, endpoint=True)

    # Evaluate lambda near the axis
    data_nae = eq.compute("lambda", grid=grid_2d_05)
    lam_nae = data_nae["lambda"]

    # Reshape to form grids on theta and phi
    zeta = (
        grid_2d_05.nodes[:, 2]
        .reshape(
            (grid_2d_05.num_theta, grid_2d_05.num_rho, grid_2d_05.num_zeta), order="F"
        )
        .squeeze()
    )

    lam_nae = lam_nae.reshape(
        (grid_2d_05.num_theta, grid_2d_05.num_rho, grid_2d_05.num_zeta), order="F"
    )

    phi = np.squeeze(zeta[0, :])
    lam_nae = np.squeeze(lam_nae[:, 0, :])

    lam_av_nae = np.mean(lam_nae, axis=0)
    np.testing.assert_allclose(
        lam_av_nae, -qic.iota * qic.nu_spline(phi), atol=1e-4, rtol=1e-2
    )

    # check |B| on axis

    grid = LinearGrid(M=eq.M_grid, N=eq.N_grid, NFP=eq.NFP, rho=np.array(1e-6))
    # Evaluate B modes near the axis
    data_nae = eq.compute(["|B|_mn", "B modes"], grid=grid)
    modes = data_nae["B modes"]
    B_mn_nae = data_nae["|B|_mn"]
    # Evaluate B on an angular grid
    theta = np.linspace(0, 2 * np.pi, 150)
    phi = np.linspace(0, 2 * np.pi, qic.nphi)
    th, ph = np.meshgrid(theta, phi)
    B_nae = np.zeros((qic.nphi, 150))

    for i, (l, m, n) in enumerate(modes):
        if m >= 0 and n >= 0:
            B_nae += B_mn_nae[i] * np.cos(m * th) * np.cos(n * ph)
        elif m >= 0 > n:
            B_nae += -B_mn_nae[i] * np.cos(m * th) * np.sin(n * ph)
        elif m < 0 <= n:
            B_nae += -B_mn_nae[i] * np.sin(m * th) * np.cos(n * ph)
        elif m < 0 and n < 0:
            B_nae += B_mn_nae[i] * np.sin(m * th) * np.sin(n * ph)
    # Eliminate the poloidal angle to focus on the toroidal behavior
    B_av_nae = np.mean(B_nae, axis=1)
    np.testing.assert_allclose(B_av_nae, np.ones(np.size(phi)) * qic.B0, atol=2e-2)


class TestGetExample:
    """Tests for desc.examples.get."""

    @pytest.mark.unit
    def test_missing_example(self):
        """Test for correct error thrown when no example is found."""
        with pytest.raises(ValueError, match="example FOO not found"):
            desc.examples.get("FOO")

    @pytest.mark.unit
    def test_example_get_eq(self):
        """Test getting a single equilibrium."""
        eq = desc.examples.get("SOLOVEV")
        assert eq.Psi == 1

    @pytest.mark.unit
    def test_example_get_eqf(self):
        """Test getting full equilibria family."""
        eqf = desc.examples.get("DSHAPE", "all")
        np.testing.assert_allclose(eqf[0].pressure.params, 0)

    @pytest.mark.unit
    def test_example_get_boundary(self):
        """Test getting boundary surface."""
        surf = desc.examples.get("HELIOTRON", "boundary")
        np.testing.assert_allclose(surf.R_lmn[surf.R_basis.get_idx(0, 1, 1)], -0.3)

    @pytest.mark.unit
    def test_example_get_pressure(self):
        """Test getting pressure profile."""
        pres = desc.examples.get("ATF", "pressure")
        np.testing.assert_allclose(pres.params[:5], [5e5, -1e6, 5e5, 0, 0])

    @pytest.mark.unit
    def test_example_get_iota(self):
        """Test getting iota profile."""
        iota = desc.examples.get("W7-X", "iota")
        np.testing.assert_allclose(
            iota.params[:5],
            [
                -8.56047021e-01,
                -3.88095412e-02,
                -6.86795128e-02,
                -1.86970315e-02,
                1.90561179e-02,
            ],
        )

    @pytest.mark.unit
    def test_example_get_current(self):
        """Test getting current profile."""
        current = desc.examples.get("NCSX", "current")
        np.testing.assert_allclose(
            current.params[:11],
            [
                0.00000000e00,
                -5.30230329e03,
                -4.65196499e05,
                2.31960013e06,
                -1.20570566e07,
                4.17520547e07,
                -9.51373229e07,
                1.38268651e08,
                -1.23703891e08,
                6.24782996e07,
                -1.36284423e07,
            ],
        )



"""Tests for _Configuration base class."""

import warnings

import numpy as np
import pytest

from desc.backend import put
from desc.equilibrium import EquilibriaFamily, Equilibrium
from desc.equilibrium.initial_guess import _initial_guess_surface
from desc.geometry import (
    FourierRZCurve,
    FourierRZToroidalSurface,
    ZernikeRZToroidalSection,
)
from desc.grid import ConcentricGrid, LinearGrid, QuadratureGrid
from desc.profiles import PowerSeriesProfile, SplineProfile


class TestConstructor:
    """Tests for creating equilibrium objects."""

    @pytest.mark.unit
    def test_defaults(self):
        """Tests that default attribute values get set correctly."""
        eq = Equilibrium()

        assert eq.spectral_indexing == "ansi"
        assert eq.NFP == 1
        assert eq.L == 1
        assert eq.M == 1
        assert eq.N == 0
        assert eq.sym is False
        assert eq.surface.eq(FourierRZToroidalSurface(sym=False))
        assert isinstance(eq.pressure, PowerSeriesProfile)
        assert isinstance(eq.current, PowerSeriesProfile)
        np.testing.assert_allclose(eq.p_l, [0])
        np.testing.assert_allclose(eq.c_l, [0])

    @pytest.mark.unit
    def test_supplied_objects(self):
        """Tests that profile and surface objects are parsed correctly."""
        pressure = SplineProfile([1, 2, 3])
        iota = SplineProfile([2, 3, 4])
        surface = FourierRZToroidalSurface(NFP=2, sym=False)
        axis = FourierRZCurve([-0.2, 10, 0.3], [0.3, 0, -0.2], NFP=2, sym=False)
        eq = Equilibrium(
            M=2,
            pressure=pressure,
            iota=iota,
            surface=surface,
            axis=axis,
            N=1,
            sym=False,
        )

        assert eq.pressure.eq(pressure)
        assert eq.iota.eq(iota)
        assert eq.spectral_indexing == "ansi"
        assert eq.NFP == 2
        assert eq.axis.NFP == 2

        np.testing.assert_allclose(axis.R_n, eq.Ra_n)
        np.testing.assert_allclose(axis.Z_n, eq.Za_n)

        surface2 = ZernikeRZToroidalSection(spectral_indexing="ansi")
        eq2 = Equilibrium(surface=surface2)
        assert eq2.surface.eq(surface2)

        surface3 = FourierRZToroidalSurface(NFP=3)
        eq3 = Equilibrium(surface=surface3)
        assert eq3.NFP == 3
        assert eq3.axis.NFP == 3

        eq4 = Equilibrium(surface=surface2, axis=None)
        np.testing.assert_allclose(eq4.axis.R_n, [10])

    @pytest.mark.unit
    def test_dict(self):
        """Test creating an equilibrium from a dictionary of arrays."""
        inputs = {
            "L": 4,
            "M": 2,
            "N": 2,
            "NFP": 3,
            "sym": False,
            "spectral_indexing": "ansi",
            "surface": np.array(
                [
                    [0, 0, 0, 10, 0],
                    [0, 1, 0, 1, 0],
                    [0, -1, 0, 0, -1],
                    [0, -1, 1, 0.1, 0.1],
                ]
            ),
            "axis": np.array([[0, 10, 0]]),
            "pressure": np.array([[0, 10], [2, 5]]),
            "iota": np.array([[0, 1], [2, 3]]),
        }
        eq = Equilibrium(**inputs)

        assert eq.L == 4
        assert eq.M == 2
        assert eq.N == 2
        assert eq.NFP == 3
        assert eq.spectral_indexing == "ansi"
        np.testing.assert_allclose(eq.p_l, [10, 5, 0])
        np.testing.assert_allclose(eq.i_l, [1, 3, 0])
        assert isinstance(eq.surface, FourierRZToroidalSurface)
        np.testing.assert_allclose(
            eq.Rb_lmn,
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                10.0,
                1.0,
                0.0,
                0.0,
                0.1,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ],
        )
        np.testing.assert_allclose(
            eq.Zb_lmn,
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                -1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.1,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ],
        )

        inputs["surface"] = np.array(
            [
                [0, 0, 0, 10, 0],
                [1, 1, 0, 1, 0.1],
                [1, -1, 0, 0.2, -1],
            ]
        )

        eq = Equilibrium(**inputs)
        assert eq.bdry_mode == "poincare"
        np.testing.assert_allclose(
            eq.Rb_lmn, [10.0, 0.2, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        )
        np.testing.assert_allclose(
            eq.Zb_lmn, [0.0, -1.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        )

    @pytest.mark.unit
    def test_asserts(self):
        """Test error checking in equilibrium creation."""
        with pytest.raises(AssertionError):
            eq = Equilibrium(L=3.4)
        with pytest.raises(AssertionError):
            eq = Equilibrium(M=3.4)
        with pytest.raises(AssertionError):
            eq = Equilibrium(N=3.4)
        with pytest.raises(ValueError):
            eq = Equilibrium(NFP=3.4j)
        with pytest.raises(ValueError):
            eq = Equilibrium(surface=np.array([[1, 1, 1, 10, 2]]))
        with pytest.raises(TypeError):
            eq = Equilibrium(surface=FourierRZCurve())
        with pytest.raises(TypeError):
            eq = Equilibrium(axis=2)
        with pytest.raises(ValueError):
            eq = Equilibrium(surface=FourierRZToroidalSurface(NFP=1), NFP=2)
        with pytest.raises(TypeError):
            eq = Equilibrium(pressure="abc")
        with pytest.raises(TypeError):
            eq = Equilibrium(iota="def")
        with pytest.raises(TypeError):
            eq = Equilibrium(current="def")
        with pytest.raises(ValueError):  # change to TypeError if allow both
            eq = Equilibrium(iota="def", current="def")
        with pytest.raises(ValueError):
            eq = Equilibrium(iota=None)
            eq.i_l = None
        with pytest.raises(ValueError):
            eq = Equilibrium(iota=PowerSeriesProfile(params=[1, 3], modes=[0, 2]))
            eq.c_l = None

    @pytest.mark.unit
    def test_supplied_coeffs(self):
        """Test passing in specific spectral coefficients."""
        R_lmn = np.random.random(3)
        Z_lmn = np.random.random(3)
        L_lmn = np.random.random(3)
        eq = Equilibrium(R_lmn=R_lmn, Z_lmn=Z_lmn, L_lmn=L_lmn, check_orientation=False)
        np.testing.assert_allclose(R_lmn, eq.R_lmn)
        np.testing.assert_allclose(Z_lmn, eq.Z_lmn)
        np.testing.assert_allclose(L_lmn, eq.L_lmn)

        with pytest.raises(ValueError):
            eq = Equilibrium(L=4, R_lmn=R_lmn)


class TestInitialGuess:
    """Tests for setting initial guess for an equilibrium."""

    @pytest.mark.unit
    def test_default_set(self):
        """Test the default initial guess."""
        eq = Equilibrium()
        eq.set_initial_guess()
        np.testing.assert_allclose(eq.compute("V")["V"], 2 * 10 * np.pi * np.pi * 1 * 1)
        del eq._axis
        eq.set_initial_guess()
        np.testing.assert_allclose(eq.compute("V")["V"], 2 * 10 * np.pi * np.pi * 1 * 1)

    @pytest.mark.unit
    def test_errors(self):
        """Test error checking for setting initial guess."""
        eq = Equilibrium()
        with pytest.raises(ValueError):
            eq.set_initial_guess(1, "a", 4, 5, 6)
        with pytest.raises(ValueError):
            eq.set_initial_guess(1, 2)
        with pytest.raises(ValueError):
            eq.set_initial_guess(eq, eq.surface)
        with pytest.raises(TypeError):
            eq.set_initial_guess(eq.surface, [1, 2, 3])
        del eq._surface
        with pytest.raises(ValueError):
            eq.set_initial_guess()

        with pytest.raises(ValueError):
            eq.set_initial_guess("path", 3)
        with pytest.raises(ValueError):
            eq.set_initial_guess("path", "hdf5")
        with pytest.raises(ValueError):
            eq.surface = eq.get_surface_at(rho=1)
            eq.change_resolution(2, 2, 2)
            _ = _initial_guess_surface(eq.R_basis, eq.R_lmn, eq.R_basis)
        with pytest.raises(ValueError):
            _ = _initial_guess_surface(
                eq.R_basis, eq.surface.R_lmn, eq.surface.R_basis, mode="foo"
            )

    @pytest.mark.unit
    def test_guess_from_other(self):
        """Test using one equilibrium as the initial guess for another."""
        eq1 = Equilibrium(L=4, M=2)
        eq2 = Equilibrium(L=2, M=1)
        eq2.set_initial_guess(eq1)

        eq2.change_resolution(L=4, M=2)
        np.testing.assert_allclose(eq1.R_lmn, eq2.R_lmn)
        np.testing.assert_allclose(eq1.Z_lmn, eq2.Z_lmn)

    @pytest.mark.unit
    def test_guess_from_surface(self):
        """Test initial guess by scaling boundary surface."""
        eq = Equilibrium()
        surface = FourierRZToroidalSurface()
        # turn the circular cross-section into an ellipse w AR=2
        surface.set_coeffs(m=-1, n=0, R=None, Z=-2)
        # move z axis up to 0.5 for no good reason
        axis = FourierRZCurve([0, 10, 0], [0, 0.5, 0])
        eq.set_initial_guess(surface, axis)
        np.testing.assert_allclose(eq.compute("V")["V"], 2 * 10 * np.pi * np.pi * 2 * 1)

    @pytest.mark.unit
    def test_guess_from_surface2(self):
        """Test initial guess by scaling interior surface."""
        eq = Equilibrium()
        # specify an interior flux surface
        surface = FourierRZToroidalSurface(rho=0.5)
        eq.set_initial_guess(surface)
        np.testing.assert_allclose(
            eq.compute("V")["V"], 2 * 10 * np.pi * np.pi * 2**2
        )

    @pytest.mark.unit
    def test_guess_from_points(self):
        """Test initial guess by fitting R,Z at specified points."""
        eq = Equilibrium(L=3, M=3, N=1)
        # these are just the default circular tokamak with a random normal
        # perturbation with std=0.03, fixed for repeatability
        eq.R_lmn = np.array(
            [
                3.94803875e-02,
                7.27321367e-03,
                -8.88095373e-03,
                1.47523628e-02,
                1.18518478e-02,
                -2.61657165e-02,
                -1.27473081e-02,
                3.26441003e-02,
                4.47427817e-03,
                1.24734770e-02,
                9.99231496e00,
                -2.74400311e-03,
                1.00447777e00,
                3.22285107e-02,
                1.16571026e-02,
                -3.15868165e-03,
                -6.77657739e-04,
                -1.97894171e-02,
                2.13535622e-02,
                -2.19703593e-02,
                5.15586341e-02,
                3.39651128e-02,
                -1.66077603e-02,
                -2.20514583e-02,
                -3.13335598e-02,
                7.16090760e-02,
                -1.30064709e-03,
                -4.00687024e-02,
                5.25583677e-02,
                4.04325991e-03,
            ]
        )
        eq.Z_lmn = np.array(
            [
                2.58179465e-02,
                -6.58108612e-03,
                3.67459870e-02,
                9.32236734e-04,
                -2.07982449e-03,
                -1.67700140e-02,
                2.56951390e-02,
                -4.49230035e-04,
                9.93325894e-02,
                4.28162330e-03,
                9.39812383e-03,
                9.95829268e-01,
                4.14468984e-02,
                -3.10725101e-02,
                -1.42026152e-02,
                -2.20423483e-02,
                -1.37389716e-02,
                -1.31592276e-02,
                -3.13922472e-02,
                1.88145630e-03,
                2.72255620e-02,
                -9.42746650e-03,
                2.15264372e-02,
                2.43549268e-02,
                5.33383228e-02,
                1.65948808e-02,
                1.45908076e-03,
                1.85101895e-02,
                1.25967662e-02,
                -2.07374046e-02,
            ]
        )
        grid = ConcentricGrid(L=6, M=6, N=2, node_pattern="ocs")
        coords = eq.compute(["R", "Z", "lambda"], grid=grid)
        eq2 = Equilibrium(L=3, M=3, N=1)
        eq2.set_initial_guess(grid.nodes, coords["R"], coords["Z"], coords["lambda"])
        np.testing.assert_allclose(eq.R_lmn, eq2.R_lmn, atol=1e-8)
        np.testing.assert_allclose(eq.Z_lmn, eq2.Z_lmn, atol=1e-8)
        np.testing.assert_allclose(eq.L_lmn, eq2.L_lmn, atol=1e-8)

    @pytest.mark.unit
    def test_NFP_error(self):
        """Check for ValueError when eq, axis, and surface NFPs do not agree."""
        axis = FourierRZCurve([-1, 10, 1], [1, 0, -1], NFP=2)
        surface2 = FourierRZToroidalSurface(NFP=2)
        surface3 = FourierRZToroidalSurface(NFP=3)

        # test axis and eq NFP not agreeing
        with pytest.raises(ValueError):
            _ = Equilibrium(surface=surface3, axis=axis, NFP=3)

        # test axis and surface NFP not agreeing
        with pytest.raises(ValueError):
            _ = Equilibrium(surface=surface2, axis=axis, NFP=3)

        # test surface and eq NFP not agreeing
        with pytest.raises(ValueError):
            _ = Equilibrium(surface=surface3, axis=axis, NFP=2)

    @pytest.mark.unit
    @pytest.mark.solve
    def test_guess_from_file(self, DSHAPE_current):
        """Test setting initial guess from saved equilibrium file."""
        path = DSHAPE_current["desc_h5_path"]
        eq1 = Equilibrium(M=13, sym=True, spectral_indexing="fringe")
        eq1.set_initial_guess(path)
        eq2 = EquilibriaFamily.load(path)[-1]

        np.testing.assert_allclose(eq1.R_lmn, eq2.R_lmn)
        np.testing.assert_allclose(eq1.Z_lmn, eq2.Z_lmn)


class TestGetSurfaces:
    """Tests for get_surface method."""

    @pytest.mark.unit
    def test_get_rho_surface(self):
        """Test getting a constant rho surface."""
        eq = Equilibrium()
        R0 = 10
        rho = 0.5
        surf = eq.get_surface_at(rho=rho)
        assert surf.rho == rho
        np.testing.assert_allclose(surf.compute("S")["S"], 4 * np.pi**2 * R0 * rho)

    @pytest.mark.unit
    def test_get_zeta_surface(self):
        """Test getting a constant zeta surface."""
        eq = Equilibrium()
        surf = eq.get_surface_at(zeta=np.pi)
        assert surf.zeta == np.pi
        rho = 1
        np.testing.assert_allclose(surf.compute("A")["A"], np.pi * rho**2)

    @pytest.mark.unit
    def test_get_theta_surface(self):
        """Test that getting a constant theta surface doesn't work yet."""
        eq = Equilibrium()
        with pytest.raises(NotImplementedError):
            _ = eq.get_surface_at(theta=np.pi)

    @pytest.mark.unit
    def test_asserts(self):
        """Test error checking in get_surface method."""
        eq = Equilibrium()
        with pytest.raises(ValueError):
            _ = eq.get_surface_at(rho=1, zeta=2)
        with pytest.raises(AssertionError):
            _ = eq.get_surface_at(rho=1.2)


@pytest.mark.unit
@pytest.mark.solve
def test_magnetic_axis(HELIOTRON_vac):
    """Test that Configuration.axis returns the true axis location."""
    eq = EquilibriaFamily.load(load_from=str(HELIOTRON_vac["desc_h5_path"]))[-1]
    axis = eq.axis
    grid = LinearGrid(N=3 * eq.N_grid, NFP=eq.NFP, rho=np.array(0.0))

    data = eq.compute(["R", "Z"], grid=grid)
    coords = axis.compute("x", grid=grid)["x"]

    np.testing.assert_allclose(coords[:, 0], data["R"])
    np.testing.assert_allclose(coords[:, 2], data["Z"])


@pytest.mark.unit
def test_is_nested():
    """Test that jacobian sign indicates whether surfaces are nested."""
    eq = Equilibrium()
    grid = QuadratureGrid(L=10, M=10, N=0)
    assert eq.is_nested(grid=grid)

    eq.change_resolution(L=2, M=2)
    eq.R_lmn = put(eq.R_lmn, eq.R_basis.get_idx(L=1, M=1, N=0), 1)
    # make unnested by setting higher order mode to same amplitude as lower order mode
    eq.R_lmn = put(eq.R_lmn, eq.R_basis.get_idx(L=2, M=2, N=0), 1)

    assert not eq.is_nested(grid=grid)
    with pytest.warns(Warning) as record:
        assert not eq.is_nested(grid=grid, msg="auto")
    assert len(record) == 1
    assert "Automatic" in str(record[0].message)
    with pytest.warns(Warning) as record:
        assert not eq.is_nested(grid=grid, msg="manual")
    assert len(record) == 1
    assert "perturbation" in str(record[0].message)
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert not eq.is_nested(grid=grid, msg=None)


@pytest.mark.unit
def test_is_nested_theta():
    """Test that new version of is nested also catches messed up theta contours."""
    eq = Equilibrium(L=6, M=6, N=0, iota=1)
    # just mess with lambda, so rho contours are the same
    eq.L_lmn += 1e-1 * np.random.default_rng(seed=3).random(eq.L_lmn.shape)
    grid = QuadratureGrid(10, 10, 0, NFP=eq.NFP)
    g1 = eq.compute("sqrt(g)", grid=grid)["sqrt(g)"]
    g2 = eq.compute("sqrt(g)_PEST", grid=grid)["sqrt(g)_PEST"]
    assert np.all(g1 > 0)  # regular jacobian will still be fine
    assert np.any(g2 < 0)  # PEST jacobian should be negative
    assert not eq.is_nested()


@pytest.mark.unit
@pytest.mark.solve
def test_get_profile(DSHAPE_current):
    """Test getting/setting iota and current profiles."""
    eq = EquilibriaFamily.load(load_from=str(DSHAPE_current["desc_h5_path"]))[-1]
    current0 = eq.current
    iota1 = eq.get_profile("iota")
    current1 = eq.get_profile("current")
    eq._current = None
    eq._iota = iota1
    iota2 = eq.get_profile("iota")
    current2 = eq.get_profile("current")

    np.testing.assert_allclose(iota1.params, iota2.params)
    np.testing.assert_allclose(current1.params, current2.params)
    x = np.linspace(0, 1, 20)
    np.testing.assert_allclose(current2(x), current0(x), rtol=1e-6, atol=1e-1)


@pytest.mark.unit
def test_kinetic_errors():
    """Test that we can't set nonexistent profile values."""
    eqp = Equilibrium(L=3, M=3, N=3, pressure=np.array([1, 0, -1]))
    eqk = Equilibrium(
        L=3,
        M=3,
        N=3,
        electron_temperature=np.array([1, 0, -1]),
        electron_density=np.array([2, 0, -2]),
    )
    params = np.arange(3)
    with pytest.raises(ValueError):
        eqk.p_l = params
    with pytest.raises(ValueError):
        eqp.Te_l = params
    with pytest.raises(ValueError):
        eqp.ne_l = params
    with pytest.raises(ValueError):
        eqp.Ti_l = params
    with pytest.raises(ValueError):
        eqp.Zeff_l = params

    params = np.ones((3, 4))
    profile = PowerSeriesProfile()
    eqk.pressure = profile
    eqp.electron_temperature = profile
    eqp.electron_density = profile
    eqp.ion_temperature = profile
    eqp.atomic_number = profile
    with pytest.raises(TypeError):
        eqk.pressure = params
    with pytest.raises(TypeError):
        eqp.electron_temperature = params
    with pytest.raises(TypeError):
        eqp.electron_density = params
    with pytest.raises(TypeError):
        eqp.ion_temperature = params
    with pytest.raises(TypeError):
        eqp.atomic_number = params

    with pytest.raises(ValueError):
        _ = Equilibrium(pressure=1, electron_density=1, electron_temperature=1)
    with pytest.raises(ValueError):
        _ = Equilibrium(electron_temperature=1)



"""Example script for recreating the precise QH configuration of Landreman and Paul."""

from desc import set_device

set_device("gpu")

import numpy as np

from desc.continuation import solve_continuation_automatic
from desc.equilibrium import EquilibriaFamily, Equilibrium
from desc.geometry import FourierRZToroidalSurface
from desc.grid import LinearGrid
from desc.objectives import (
    AspectRatio,
    FixBoundaryR,
    FixBoundaryZ,
    FixCurrent,
    FixPressure,
    FixPsi,
    ForceBalance,
    ObjectiveFunction,
    QuasisymmetryTwoTerm,
)
from desc.optimize import Optimizer

# create initial equilibrium
surf = FourierRZToroidalSurface(
    R_lmn=[1, 0.125, 0.3],
    Z_lmn=[-0.125, -0.3],
    modes_R=[[0, 0], [1, 0], [0, 1]],
    modes_Z=[[-1, 0], [0, -1]],
    NFP=4,
)
eq = Equilibrium(M=8, N=8, Psi=0.04, surface=surf)
eq = solve_continuation_automatic(eq, objective="force", verbose=3)[-1]
eqfam = EquilibriaFamily(eq)

# optimize in steps
grid = LinearGrid(M=eq.M, N=eq.N, NFP=eq.NFP, rho=np.array([0.6, 0.8, 1.0]), sym=True)
for n in range(1, eq.M + 1):
    print("\n==================================")
    print("Optimizing boundary modes M,N <= {}".format(n))
    print("====================================")
    objective = ObjectiveFunction(
        (
            QuasisymmetryTwoTerm(
                eq=eqfam[-1], helicity=(1, eq.NFP), grid=grid, normalize=False
            ),
            AspectRatio(eq=eqfam[-1], target=8, weight=1e1, normalize=False),
        ),
        verbose=0,
    )
    R_modes = np.vstack(
        (
            [0, 0, 0],
            eq.surface.R_basis.modes[
                np.max(np.abs(eq.surface.R_basis.modes), 1) > n, :
            ],
        )
    )
    Z_modes = eq.surface.Z_basis.modes[
        np.max(np.abs(eq.surface.Z_basis.modes), 1) > n, :
    ]
    constraints = (
        ForceBalance(eq=eqfam[-1]),  # J x B - grad(p) = 0
        FixBoundaryR(eq=eqfam[-1], modes=R_modes),
        FixBoundaryZ(eq=eqfam[-1], modes=Z_modes),
        FixPressure(eq=eqfam[-1]),
        FixCurrent(eq=eqfam[-1]),
        FixPsi(eq=eqfam[-1]),
    )
    optimizer = Optimizer("proximal-lsq-exact")
    eq_new, out = eqfam[-1].optimize(
        objective=objective,
        constraints=constraints,
        optimizer=optimizer,
        maxiter=50,
        verbose=3,
        copy=True,
        options={
            "initial_trust_radius": 0.5,
            "perturb_options": {"verbose": 0},
            "solve_options": {"verbose": 0},
        },
    )
    eqfam.append(eq_new)

eqfam.save("precise_QH_output.h5")



#!/usr/bin/env python3
"""
Run this script to regenerate all the DESC equilibria in desc/examples.

python3 regenerate_all_equilibria.py

Copy the two lines below to the main code to run it on a GPU
from desc import set_device
set_device("gpu")
"""
import glob
import os
import subprocess as spr

from desc.__main__ import main
from desc.io import load
from desc.vmec import VMECIO

pwd = os.getcwd()

for fname in glob.glob(pwd + "/*.h5"):
    if (
        fname.split(".")[-1] == "h5"
        and os.path.isfile(fname.split(".")[-1] + ".nc") is True
        and fname.split(".")[0].split("/")[-1].split("_")[0] != "precise"
    ):
        finputname = fname.split(".")[0].split("/")[-1].split("_")[0]
        print(f"Running the input file {finputname} \n")
        main(cl_args=[str(f"{finputname}"), "-vv"])
        # save wout file
        eq = load(f"{pwd}/{finputname}")[-1]
        VMECIO.save(eq, "wout_" + fname + ".nc", surfs=256)
    elif (
        fname.split(".")[-1] == "h5"
        and os.path.isfile(fname + ".nc") is False
        and fname.split(".")[0].split("/")[-1].split("_")[0] != "precise"
    ):
        finputname = fname.split(".")[0].split("/")[-1].split("_")[0]
        print(f"Running the input file {finputname} \n")
        main(cl_args=[str(f"{finputname}"), "-vv"])
    else:
        continue

main(cl_args=["DSHAPE_CURRENT", "-vv"])
spr.call(["python3 -u precise_QA.py"], shell=True)
spr.call(["python3 -u precise_QH.py"], shell=True)



"""Tests for computing iota from fixed current profile and vice versa."""

import numpy as np
import pytest
from tests.test_axis_limits import not_implemented_limits

import desc.io
from desc.compute import compute as compute_fun
from desc.compute import data_index, get_params, get_profiles, get_transforms
from desc.grid import ConcentricGrid, LinearGrid, QuadratureGrid


class _ExactValueProfile:
    """Monkey patches the compute method of desc.Profile for testing."""

    def __init__(self, eq, grid, profile_name):
        names = []
        dr = 0
        while (
            profile_name + "_" * bool(dr) + "r" * dr
            in data_index["desc.equilibrium.equilibrium.Equilibrium"]
        ):
            names += [profile_name + "_" * bool(dr) + "r" * dr]
            dr += 1
        data = compute_fun(
            parameterization="desc.equilibrium.equilibrium.Equilibrium",
            names=names,
            params=get_params(keys=names, obj=eq, has_axis=grid.axis.size),
            transforms=get_transforms(keys=names, obj=eq, grid=grid),
            profiles=get_profiles(keys=names, obj=eq, grid=grid),
        )
        self.output = {dr: data[name] for dr, name in enumerate(names)}

    def compute(self, grid, params, dr, *args, **kwargs):
        return self.output.get(dr, np.nan)


class TestConstrainCurrent:
    """Tests for computing iota from fixed current profile and vice versa.

    The equations referenced in the descriptions below refer to the document
    attached to the description of GitHub pull request #556.
    """

    @pytest.mark.unit
    @pytest.mark.solve
    def test_iota_to_current_and_back(self, DSHAPE):
        """Test we can recover the rotational transform in simple cases.

        Given ``iota``, compute ``I(iota)`` as defined in equation 6.
        Use that ``I(iota)`` to compute ``iota(I)`` as defined in equation 18.
        Check that this output matches the input.

        Cross-validates the computation of the rotational transform by feeding
        the resulting enclosed net toroidal current of an equilibrium with a
        known rotational transform to a method which computes the rotational
        transform as a function of enclosed net toroidal current.

        This tests that rotational transform computations from known toroidal
        current profiles are correct, among other things. For example, this test
        will fail if the compute functions for iota given a current profile are
        incorrect.
        """
        names = ["iota", "iota_r"]
        while names[-1] + "r" in data_index["desc.equilibrium.equilibrium.Equilibrium"]:
            names += [names[-1] + "r"]

        def test(stellarator, grid_type):
            eq = desc.io.load(load_from=str(stellarator["desc_h5_path"]))[-1]
            kwargs = {"L": eq.L_grid, "M": eq.M_grid, "N": eq.N_grid, "NFP": eq.NFP}
            if grid_type != QuadratureGrid:
                kwargs["sym"] = eq.sym
            if grid_type == LinearGrid:
                kwargs["axis"] = True
            grid = grid_type(**kwargs)

            params = get_params(keys=names, obj=eq, has_axis=grid.axis.size)
            transforms = get_transforms(keys=names, obj=eq, grid=grid)
            profiles = get_profiles(keys=names, obj=eq, grid=grid)
            # Compute rotational transform directly from the power series which
            # defines it.
            desired = {
                name: profiles[names[0]].compute(grid, params=params["i_l"], dr=dr)
                for dr, name in enumerate(names)
            }
            # Compute rotational transform using the below toroidal current,
            # which is monkey patched to return equation 6 in amps.
            monkey_patched_profile = _ExactValueProfile(
                eq=eq, grid=grid, profile_name="current"
            )
            actual = compute_fun(
                parameterization="desc.equilibrium.equilibrium.Equilibrium",
                names=names,
                params=dict(params, c_l=None, i_l=None),
                transforms=transforms,
                profiles=dict(profiles, current=monkey_patched_profile, iota=None),
            )

            # Evaluating the order n derivative of the rotational transform at
            # the magnetic axis using a given toroidal current profile requires
            # the order n + 2 derivative of that toroidal current profile.
            # Computing the order n derivative of the resulting toroidal current
            # as a function of rotational transform requires the order n
            # derivative of the rotational transform.
            # Therefore, this test cannot be used to cross-validate the magnetic
            # axis limit of the two highest order derivatives of the rotational
            # transform in data_index.
            at_axis = grid.nodes[:, 0] == 0
            for index, name in enumerate(names):
                # don't try to validate if we can't compute axis limit yet
                validate_axis = (
                    index < len(names) - 2
                ) and name not in not_implemented_limits
                mask = (validate_axis & at_axis) | ~at_axis
                np.testing.assert_allclose(
                    actual=actual[name][mask],
                    desired=desired[name][mask],
                    equal_nan=False,
                    err_msg=name,
                )

        # Only makes sense to test on configurations with fixed iota profiles.
        test(DSHAPE, QuadratureGrid)
        test(DSHAPE, ConcentricGrid)
        test(DSHAPE, LinearGrid)

    @pytest.mark.unit
    @pytest.mark.solve
    def test_current_to_iota_and_back(self, HELIOTRON_vac):
        """Test we can recover the enclosed net toroidal current in simple cases.

        Given ``I``, compute ``iota(I)`` as defined in equation 18.
        Use that ``iota(I)`` to compute ``I(iota)`` as defined in equation 6.
        Check that this output matches the input.

        Cross-validates the computation of the enclosed net toroidal current by
        feeding the resulting rotational transform of an equilibrium with a
        known toroidal current profile to a method which computes the enclosed
        net toroidal current as a function of rotational transform.

        This tests that toroidal current computations from known rotational
        transform profiles are correct, among other things. For example, this
        test will fail if the compute functions for current as a function of
        iota are incorrect.
        """
        names = ["current", "current_r"]
        while names[-1] + "r" in data_index["desc.equilibrium.equilibrium.Equilibrium"]:
            names += [names[-1] + "r"]

        def test(stellarator, grid_type):
            eq = desc.io.load(load_from=str(stellarator["desc_h5_path"]))[-1]
            kwargs = {"L": eq.L_grid, "M": eq.M_grid, "N": eq.N_grid, "NFP": eq.NFP}
            if grid_type != QuadratureGrid:
                kwargs["sym"] = eq.sym
            if grid_type == LinearGrid:
                kwargs["axis"] = True
            grid = grid_type(**kwargs)

            params = get_params(keys=names, obj=eq, has_axis=grid.axis.size)
            transforms = get_transforms(keys=names, obj=eq, grid=grid)
            profiles = get_profiles(keys=names, obj=eq, grid=grid)
            # Compute toroidal current directly from the power series which
            # defines it. Recall that vacuum objective sets a current profile.
            desired = {
                name: profiles[names[0]].compute(grid, params=params["c_l"], dr=dr)
                for dr, name in enumerate(names)
            }
            # Compute toroidal current using the below rotational transform,
            # which is monkey patched to return equation 18.
            monkey_patched_profile = _ExactValueProfile(
                eq=eq, grid=grid, profile_name="iota"
            )
            actual = compute_fun(
                parameterization="desc.equilibrium.equilibrium.Equilibrium",
                names=names,
                params=dict(params, c_l=None, i_l=None),
                transforms=transforms,
                profiles=dict(profiles, current=None, iota=monkey_patched_profile),
            )

            at_axis = grid.nodes[:, 0] == 0
            for name in names:
                # don't try to validate if we can't compute axis limit yet
                validate_axis = name not in not_implemented_limits
                mask = (validate_axis & at_axis) | ~at_axis
                np.testing.assert_allclose(
                    actual=actual[name][mask],
                    desired=desired[name][mask],
                    atol=1e-7,
                    equal_nan=False,
                    err_msg=name,
                )

        # Only makes sense to test on configurations with fixed current profiles.
        test(HELIOTRON_vac, QuadratureGrid)
        test(HELIOTRON_vac, ConcentricGrid)
        test(HELIOTRON_vac, LinearGrid)
